"""
GSLCodeGenerators for code that uses the ODE solver provided by the GNU Scientific Library (GSL)
"""

import os
import re

import numpy as np

from brian2.codegen.generators import c_data_type
from brian2.codegen.permutation_analysis import (
    OrderDependenceError,
    check_for_order_independence,
)
from brian2.codegen.translation import make_statements
from brian2.core.functions import Function
from brian2.core.preferences import BrianPreference, PreferenceError, prefs
from brian2.core.variables import ArrayVariable, AuxiliaryVariable, Constant
from brian2.parsing.statements import parse_statement
from brian2.units.fundamentalunits import fail_for_dimension_mismatch
from brian2.utils.logger import get_logger
from brian2.utils.stringtools import get_identifiers, word_substitute

__all__ = ["GSLCodeGenerator", "GSLCPPCodeGenerator", "GSLCythonCodeGenerator"]


logger = get_logger(__name__)


def valid_gsl_dir(val):
    """
    Validate given string to be path containing required GSL files.
    """
    if val is None:
        return True
    if not isinstance(val, str):
        raise PreferenceError(
            f"Illegal value for GSL directory: {str(val)}, has to be str"
        )
    if not os.path.isdir(val):
        raise PreferenceError(
            f"Illegal value for GSL directory: {val}, has to be existing directory"
        )
    if any(
        not os.path.isfile(os.path.join(val, "gsl", filename))
        for filename in ["gsl_odeiv2.h", "gsl_errno.h", "gsl_matrix.h"]
    ):
        raise PreferenceError(
            f"Illegal value for GSL directory: '{val}', "
            "has to contain gsl_odeiv2.h, gsl_errno.h "
            "and gsl_matrix.h"
        )
    return True


prefs.register_preferences(
    "GSL",
    "Directory containing GSL code",
    directory=BrianPreference(
        validator=valid_gsl_dir,
        docs=(
            "Set path to directory containing GSL header files (gsl_odeiv2.h etc.)"
            "\nIf this directory is already in Python's include (e.g. because of "
            "conda installation), this path can be set to None."
        ),
        default=None,
    ),
)


class GSLCodeGenerator:
    """
    GSL code generator.

    Notes
    -----
    Approach is to first let the already existing code generator for a target
    language do the bulk of the translating from abstract_code to actual code.
    This generated code is slightly adapted to render it GSL compatible.
    The most critical part here is that the vector_code that is normally
    contained in a loop in the ```main()``` is moved to the function ```_GSL_func```
    that is sent to the GSL integrator. The variables used in the vector_code are
    added to a struct named ```dataholder``` and their values are set from the
    Brian namespace just before the scalar code block.
    """

    def __init__(
        self,
        variables,
        variable_indices,
        owner,
        iterate_all,
        codeobj_class,
        name,
        template_name,
        override_conditional_write=None,
        allows_scalar_write=False,
    ):
        self.generator = codeobj_class.original_generator_class(
            variables,
            variable_indices,
            owner,
            iterate_all,
            codeobj_class,
            name,
            template_name,
            override_conditional_write,
            allows_scalar_write,
        )
        self.method_options = dict(owner.state_updater.method_options)
        self.integrator = owner.state_updater.integrator
        # default timestep to start with is the timestep of the NeuronGroup itself
        self.method_options["dt_start"] = owner.dt.variable.get_value()[0]
        self.variable_flags = owner.state_updater._gsl_variable_flags

    def __getattr__(self, item):
        return getattr(self.generator, item)

    # A series of functions that should be overridden by child class:
    def c_data_type(self, dtype):
        """
        Get string version of object dtype that is attached to Brian variables. c
        pp_generator already has this function, but the Cython generator does not,
        but we need it for GSL code generation.
        """
        return NotImplementedError

    def initialize_array(self, varname, values):
        """
        Initialize a static array with given floating point values. E.g. in C++,
        when called with arguments ``array`` and ``[1.0, 3.0, 2.0]``, this
        method should return ``double array[] = {1.0, 3.0, 2.0}``.

        Parameters
        ----------
        varname : str
            The name of the array variable that should be initialized
        values : list of float
            The values that should be assigned to the array

        Returns
        -------
        code : str
            One or more lines of array initialization code.
        """
        raise NotImplementedError

    def var_init_lhs(self, var, type):
        """
        Get string version of the left hand side of an initializing expression

        Parameters
        ----------
        var : str
        type : str

        Returns
        -------
        code : str
            For cpp returns type + var, while for cython just var
        """
        raise NotImplementedError

    def unpack_namespace_single(self, var_obj, in_vector, in_scalar):
        """
        Writes the code necessary to pull single variable out of the Brian
        namespace into the generated code.

        The code created is significantly different between cpp and cython,
        so I decided to not make this function general
        over all target languages (i.e. in contrast to most other functions
        that only have syntactical differences)
        """
        raise NotImplementedError

    # GSL functions that are the same for all target languages:
    def find_function_names(self):
        """
        Return a list of used function names in the self.variables dictionary

        Functions need to be ignored in the GSL translation process, because the
        brian generator already sufficiently
        dealt with them. However, the brian generator also removes them from the
        variables dict, so there is no
        way to check whether an identifier is a function after the brian
        translation process. This function is called
        before this translation process and the list of function names is stored
        to be used in the GSL translation.

        Returns
        -------
        function_names : list
            list of strings that are function names used in the code
        """
        variables = self.variables
        return [
            var for var, var_obj in variables.items() if isinstance(var_obj, Function)
        ]

    def is_cpp_standalone(self):
        """
        Check whether we're running with cpp_standalone.

        Test if `get_device()` is instance `CPPStandaloneDevice`.

        Returns
        -------
        is_cpp_standalone : bool
            whether currently using cpp_standalone device

        See Also
        --------
        is_constant_and_cpp_standalone : uses the returned value
        """
        # imports here to avoid circular imports
        from brian2.devices.cpp_standalone.device import CPPStandaloneDevice
        from brian2.devices.device import get_device

        device = get_device()
        return isinstance(device, CPPStandaloneDevice)

    def is_constant_and_cpp_standalone(self, var_obj):
        """Check whether self.cpp_standalone and variable is Constant.

        This check is needed because in the case of using the cpp_standalone device we do not
        want to apply our GSL variable conversion (var --> _GSL_dataholder.var), because the cpp_standalone
        code generation process involves replacing constants with their actual value ('freezing').
        This results in code that looks like (if for example var = 1.2): _GSL_dataholder.1.2 = 1.2 and _GSL_dataholder->1.2.
        To prevent repetitive calls to get_device() etc. the outcome of is_cpp_standalone is saved.

        Parameters
        ----------
        var_obj : `Variable`
            instance of brian Variable class describing the variable

        Returns
        -------
        is_cpp_standalone : bool
            whether the used device is cpp_standalone and the given variable is an instance of Constant
        """
        if not hasattr(self, "cpp_standalone"):
            self.cpp_standalone = self.is_cpp_standalone()
        return isinstance(var_obj, Constant) and self.cpp_standalone

    def find_differential_variables(self, code):
        """
        Find the variables that were tagged _gsl_{var}_f{ind} and return var, ind pairs.

        `GSLStateUpdater` tagged differential variables and here we extract the information given in these tags.

        Parameters
        ----------
        code : list of strings
            A list of strings containing gsl tagged variables

        Returns
        -------
        diff_vars : dict
            A dictionary with variable names as keys and differential equation index as value
        """
        diff_vars = {}
        for expr_set in code:
            for expr in expr_set.split("\n"):
                expr = expr.strip(" ")
                try:
                    lhs, op, rhs, comment = parse_statement(expr)
                except ValueError:
                    pass
                m = re.search("_gsl_(.+?)_f([0-9]*)$", lhs)
                if m:
                    diff_vars[m.group(1)] = m.group(2)
        return diff_vars

    def diff_var_to_replace(self, diff_vars):
        """
        Add differential variable-related strings that need to be replaced to go
        from normal brian to GSL code

        From the code generated by Brian's 'normal' generators (cpp_generator or
        cython_generator a few bits of text need to be replaced to get GSL
        compatible code. The bits of text related to differential equation
        variables are put in the replacer dictionary in this function.

        Parameters
        ----------
        diff_vars : dict
            dictionary with variables as keys and differential equation index as value

        Returns
        -------
        to_replace : dict
            dictionary with strings that need to be replaced as keys and the
            strings that will replace them as values
        """
        variables = self.variables
        to_replace = {}
        for var, diff_num in list(diff_vars.items()):
            to_replace.update(self.var_replace_diff_var_lhs(var, diff_num))
            var_obj = variables[var]
            array_name = self.generator.get_array_name(var_obj, access_data=True)
            idx_name = "_idx"  # TODO: could be dynamic?
            replace_what = f"{var} = {array_name}[{idx_name}]"
            replace_with = f"{var} = _GSL_y[{diff_num}]"
            to_replace[replace_what] = replace_with
        return to_replace

    def get_dimension_code(self, diff_num):
        """
        Generate code for function that sets the dimension of the ODE system.

        GSL needs to know how many differential variables there are in the
        ODE system. Since the current approach is to have the code in the vector
        loop the same for all simulations, this dimension is set by an external
        function. The code for this set_dimension functon is written here.
        It is assumed the code will be the same for each target language with the
        exception of some syntactical differences

        Parameters
        ----------
        diff_num : int
            Number of differential variables that describe the ODE system

        Returns
        -------
        set_dimension_code : str
            The code describing the target language function in a single string
        """
        code = ["\n{start_declare}int set_dimension(size_t * dimension){open_function}"]
        code += ["\tdimension[0] = %d{end_statement}" % diff_num]
        code += ["\treturn GSL_SUCCESS{end_statement}{end_function}"]
        return ("\n").join(code).format(**self.syntax)

    def yvector_code(self, diff_vars):
        """
        Generate code for function dealing with GSLs y vector.

        The values of differential variables have to be transferred from
        Brian's namespace to a vector that is given to GSL. The transferring
        from Brian --> y and back from y --> Brian after integration happens in
        separate functions. The code for these is written here.

        Parameters
        ----------
        diff_vars : dictionary
            Dictionary containing variable names as keys (str) and differential
            variable index as value

        Returns
        -------
        yvector_code : str
            The code for the two functions (``_fill_y_vector`` and
            ``_empty_y_vector``) as single string.
        """
        fill_y = [
            "\n{start_declare}int _fill_y_vector(_dataholder *"
            "_GSL_dataholder, double * _GSL_y, int _idx){open_function}"
        ]
        empty_y = [
            "\n{start_declare}int _empty_y_vector(_dataholder * "
            "_GSL_dataholder, double * _GSL_y, int _idx){"
            "open_function}"
        ]
        for var, diff_num in list(diff_vars.items()):
            diff_num = int(diff_num)
            array_name = self.generator.get_array_name(
                self.variables[var], access_data=True
            )
            fill_y += [
                "\t_GSL_y[%d] = _GSL_dataholder{access_pointer}%s[_idx]{end_statement}"
                % (diff_num, array_name)
            ]
            empty_y += [
                "\t_GSL_dataholder{access_pointer}%s[_idx] = _GSL_y[%d]{end_statement}"
                % (array_name, diff_num)
            ]
        fill_y += ["\treturn GSL_SUCCESS{end_statement}{end_function}"]
        empty_y += ["\treturn GSL_SUCCESS{end_statement}{end_function}"]
        return ("\n").join(fill_y + empty_y).format(**self.syntax)

    def make_function_code(self, lines):
        """
        Add lines of GSL translated vector code to 'non-changing' _GSL_func code.

        Adds nonchanging aspects of GSL _GSL_func code to lines of code
        written somewhere else (`translate_vector_code`). Here these lines
        are put between the non-changing parts of the code and the
        target-language specific syntax is added.

        Parameters
        ----------
        lines : str
            Code containing GSL version of equations

        Returns
        -------
        function_code : str
            code describing ``_GSL_func`` that is sent to GSL integrator.
        """
        code = [
            "\n{start_declare}int _GSL_func(double t, const double "
            "_GSL_y[], double f[], void * params){open_function}"
            "\n\t{start_declare}_dataholder * _GSL_dataholder = {open_cast}"
            "_dataholder *{close_cast} params{end_statement}"
            "\n\t{start_declare}int _idx = _GSL_dataholder{access_pointer}_idx"
            "{end_statement}"
        ]
        code += [lines]
        code += ["\treturn GSL_SUCCESS{end_statement}{end_function}"]
        return ("\n").join(code).format(**self.syntax)

    def write_dataholder_single(self, var_obj):
        """
        Return string declaring a single variable in the ``_dataholder`` struct.

        Parameters
        ----------
        var_obj : `Variable`

        Returns
        -------
        code : str
            string describing this variable object as required for the ``_dataholder`` struct
            (e.g. ``double* _array_neurongroup_v``)
        """
        dtype = self.c_data_type(var_obj.dtype)
        if isinstance(var_obj, ArrayVariable):
            pointer_name = self.get_array_name(var_obj, access_data=True)
            try:
                restrict = self.generator.restrict
            except AttributeError:
                restrict = ""
            if var_obj.scalar or var_obj.size == 1:
                restrict = ""
            return f"{dtype}* {restrict} {pointer_name}{{end_statement}}"
        else:
            return f"{dtype} {var_obj.name}{{end_statement}}"

    def write_dataholder(self, variables_in_vector):
        """
        Return string with full code for _dataholder struct.

        Parameters
        ----------
        variables_in_vector : dict
            dictionary containing variable name as key and `Variable` as value

        Returns
        -------
        code : str
            code for _dataholder struct
        """
        code = ["\n{start_declare}struct _dataholder{open_struct}"]
        code += ["\tint _idx{end_statement}"]
        for var, var_obj in list(variables_in_vector.items()):
            if (
                var == "t"
                or "_gsl" in var
                or self.is_constant_and_cpp_standalone(var_obj)
            ):
                continue
            code += [f"	{self.write_dataholder_single(var_obj)}"]
        code += ["{end_struct}"]
        return ("\n").join(code).format(**self.syntax)

    def scale_array_code(self, diff_vars, method_options):
        """
        Return code for definition of ``_GSL_scale_array`` in generated code.

        Parameters
        ----------
        diff_vars : dict
            dictionary with variable name (str) as key and differential variable
            index (int) as value
        method_options : dict
            dictionary containing integrator settings

        Returns
        -------
        code : str
            full code describing a function returning a array containing doubles
            with the absolute errors for each differential variable (according
            to their assigned index in the GSL StateUpdater)
        """
        # get scale values per variable from method_options
        abs_per_var = method_options["absolute_error_per_variable"]
        abs_default = method_options["absolute_error"]

        if not isinstance(abs_default, float):
            raise TypeError(
                "The absolute_error key in method_options should be "
                f"a float. Was type {type(abs_default)}"
            )

        if abs_per_var is None:
            diff_scale = {var: float(abs_default) for var in list(diff_vars.keys())}
        elif isinstance(abs_per_var, dict):
            diff_scale = {}
            for var, error in list(abs_per_var.items()):
                # first do some checks on input
                if var not in diff_vars:
                    if var not in self.variables:
                        raise KeyError(
                            "absolute_error specified for variable that "
                            f"does not exist: {var}"
                        )
                    else:
                        raise KeyError(
                            "absolute_error specified for variable that is "
                            f"not being integrated: {var}"
                        )
                fail_for_dimension_mismatch(
                    error,
                    self.variables[var],
                    "Unit of absolute_error_per_variable "
                    f"for variable {var} does not match "
                    "unit of variable itself",
                )
                # if all these are passed we can add the value for error in base units
                diff_scale[var] = float(error)
            # set the variables that are not mentioned to default value
            for var in list(diff_vars.keys()):
                if var not in abs_per_var:
                    diff_scale[var] = float(abs_default)
        else:
            raise TypeError(
                "The absolute_error_per_variable key in method_options "
                "should either be None or a dictionary "
                "containing the error for each individual state variable. "
                f"Was type {type(abs_per_var)}"
            )
        # write code
        return self.initialize_array(
            "_GSL_scale_array", [diff_scale[var] for var in sorted(diff_vars)]
        )

    def find_undefined_variables(self, statements):
        r"""
        Find identifiers that are not in ``self.variables`` dictionary.

        Brian does not save the ``_lio_`` variables it uses anywhere. This is
        problematic for our GSL implementation because we save the lio variables
        in the ``_dataholder`` struct (for which we need the datatype of the
        variables). This function adds the left hand side variables that are
        used in the vector code to the variable dictionary as
        `AuxiliaryVariable`\ s (all we need later is the datatype).

        Parameters
        ----------
        statements : list
            list of statement objects (need to have the dtype attribute)

        Notes
        -----
        I keep ``self.variables`` and ``other_variables`` separate so I can
        distinguish what variables are in the Brian namespace and which ones are
        defined in the code itself.
        """
        variables = self.variables
        other_variables = {}
        for statement in statements:
            var = statement.var
            if var not in variables:
                other_variables[var] = AuxiliaryVariable(var, dtype=statement.dtype)
        return other_variables

    def find_used_variables(self, statements, other_variables):
        """
        Find all the variables used on the right hand side of the given
        expressions.

        Parameters
        ----------
        statements : list
            list of statement objects

        Returns
        -------
        used_variables : dict
            dictionary of variables that are used as variable name (str),
            `Variable` pairs.
        """
        variables = self.variables
        used_variables = {}
        for statement in statements:
            rhs = statement.expr
            for var in get_identifiers(rhs):
                if var in self.function_names:
                    continue
                try:
                    var_obj = variables[var]
                except KeyError:
                    var_obj = other_variables[var]
                used_variables[var] = var_obj  # save as object because this has
                # all needed info (dtype, name, isarray)

        # I don't know a nicer way to do this, the above way misses write
        # variables (e.g. not_refractory)..
        read, write, _ = self.array_read_write(statements)
        for var in read | write:
            if var not in used_variables:
                used_variables[var] = variables[var]  # will always be array and
                # thus exist in variables

        return used_variables

    def to_replace_vector_vars(self, variables_in_vector, ignore=frozenset()):
        """
        Create dictionary containing key, value pairs with to be replaced text
        to translate from conventional Brian to GSL.

        Parameters
        ----------
        variables_in_vector : dict
            dictionary with variable name (str), `Variable` pairs of variables
            occurring in vector code
        ignore : set, optional
            set of strings with variable names that should be ignored

        Returns
        -------
        to_replace : dict
            dictionary with strings that need to be replaced i.e. _lio_1 will be
            _GSL_dataholder._lio_1 (in cython) or _GSL_dataholder->_lio_1 (cpp)

        Notes
        -----
        t will always be added because GSL defines its own t.
        i.e. for cpp: {'const t = _ptr_array_defaultclock_t[0];' : ''}
        """
        access_pointer = self.syntax["access_pointer"]
        to_replace = {}
        t_in_code = None
        for var, var_obj in list(variables_in_vector.items()):
            if var_obj.name == "t":
                t_in_code = var_obj
                continue
            if "_gsl" in var or var in ignore:
                continue
            if self.is_constant_and_cpp_standalone(var_obj):
                # does not have to be processed by GSL generator
                self.variables_to_be_processed.remove(var_obj.name)
                continue
            if isinstance(var_obj, ArrayVariable):
                pointer_name = self.get_array_name(var_obj, access_data=True)
                to_replace[pointer_name] = (
                    f"_GSL_dataholder{access_pointer}{pointer_name}"
                )
            else:
                to_replace[var] = f"_GSL_dataholder{access_pointer}{var}"

        # also make sure t declaration is replaced if in code
        if t_in_code is not None:
            t_declare = self.var_init_lhs("t", "const double ")
            array_name = self.get_array_name(t_in_code, access_data=True)
            end_statement = self.syntax["end_statement"]
            replace_what = f"{t_declare} = {array_name}[0]{end_statement}"
            to_replace[replace_what] = ""
            self.variables_to_be_processed.remove("t")

        return to_replace

    def unpack_namespace(
        self, variables_in_vector, variables_in_scalar, ignore=frozenset()
    ):
        """
        Write code that unpacks Brian namespace to cython/cpp namespace.

        For vector code this means putting variables in _dataholder (i.e.
        _GSL_dataholder->var or _GSL_dataholder.var = ...)
        Note that code is written so a variable could occur both in scalar and
        vector code

        Parameters
        ----------
        variables_in_vector : dict
            dictionary with variable name (str), `Variable` pairs of variables
            occurring in vector code
        variables_in_scalar : dict
            dictionary with variable name (str), `Variable` pairs of variables
             occurring in scalar code
        ignore : set, optional
            set of string names of variables that should be ignored

        Returns
        -------
        unpack_namespace_code : str
            code fragment unpacking the Brian namespace (setting variables in
            the _dataholder struct in case of vector)
        """
        code = []
        for var, var_obj in list(self.variables.items()):
            if var in ignore:
                continue
            if self.is_constant_and_cpp_standalone(var_obj):
                continue
            in_vector = var in variables_in_vector
            in_scalar = var in variables_in_scalar
            if in_vector:
                self.variables_to_be_processed.remove(var)
            code += [self.unpack_namespace_single(var_obj, in_vector, in_scalar)]
        return ("\n").join(code)

    def translate_vector_code(self, code_lines, to_replace):
        """
        Translate vector code to GSL compatible code by substituting fragments
        of code.

        Parameters
        ----------
        code_lines : list
            list of strings describing the vector_code
        to_replace: dict
            dictionary with to be replaced strings (see to_replace_vector_vars
            and to_replace_diff_vars)

        Returns
        -------
        vector_code : str
            New code that is now to be added to the function that is sent to the
            GSL integrator
        """
        code = []
        for expr_set in code_lines:
            for line in expr_set.split(
                "\n"
            ):  # every line seperate to make tabbing correct
                code += [f"	{line}"]
        code = ("\n").join(code)
        code = word_substitute(code, to_replace)

        # special substitute because of limitations of regex word boundaries with
        # variable[_idx]
        for from_sub, to_sub in list(to_replace.items()):
            m = re.search(r"\[(\w+)\];?$", from_sub)
            if m:
                code = re.sub(re.sub(r"\[", r"\[", from_sub), to_sub, code)

        if "_gsl" in code:
            raise AssertionError(
                "Translation failed, _gsl still in code (should only "
                "be tag, and should be replaced).\n"
                f"Code:\n{code}"
            )

        return code

    def translate_scalar_code(
        self, code_lines, variables_in_scalar, variables_in_vector
    ):
        """
        Translate scalar code: if calculated variables are used in the vector_code
        their value is added to the variable in the _dataholder.

        Parameters
        ----------
        code_lines : list
            list of strings containing scalar code
        variables_in_vector : dict
            dictionary with variable name (str), `Variable` pairs of variables
            occurring in vector code
        variables_in_scalar : dict
            dictionary with variable name (str), `Variable` pairs of variables
            occurring in scalar code

        Returns
        -------
        scalar_code : str
            code fragment that should be injected in the main before the loop
        """
        code = []
        for line in code_lines:
            m = re.search(r"(\w+ = .*)", line)
            try:
                new_line = m.group(1)
                var, op, expr, comment = parse_statement(new_line)
            except (ValueError, AttributeError):
                code += [line]
                continue
            if var in list(variables_in_scalar.keys()):
                code += [line]
            elif var in list(variables_in_vector.keys()):
                if var == "t":
                    continue
                try:
                    self.variables_to_be_processed.remove(var)
                except KeyError:
                    raise AssertionError(
                        "Trying to process variable named %s by "
                        "putting its value in the _GSL_dataholder "
                        "based on scalar code, but the variable "
                        "has been processed already." % var
                    )
                code += [f"_GSL_dataholder.{var} {op} {expr} {comment}"]
        return "\n".join(code)

    def add_gsl_variables_as_non_scalar(self, diff_vars):
        """
        Add _gsl variables as non-scalar.

        In `GSLStateUpdater` the differential equation variables are substituted
        with GSL tags that describe the information needed to translate the
        conventional Brian code to GSL compatible code. This function tells
        Brian that the variables that contain these tags should always be vector
        variables. If we don't do this, Brian renders the tag-variables as
        scalar if no vector variables are used in the right hand side of the
        expression.

        Parameters
        ----------
        diff_vars : dict
            dictionary with variables as keys and differential equation index as
            value
        """
        for var, ind in list(diff_vars.items()):
            name = f"_gsl_{var}_f{ind}"
            self.variables[name] = AuxiliaryVariable(var, scalar=False)

    def add_meta_variables(self, options):
        if options["use_last_timestep"]:
            try:
                N = self.variables["N"].item()
                self.owner.variables.add_array(
                    "_last_timestep",
                    size=N,
                    values=np.ones(N) * options["dt_start"],
                    dtype=np.float64,
                )
            except KeyError:
                # has already been run
                pass
            self.variables["_last_timestep"] = self.owner.variables.get(
                "_last_timestep"
            )
            pointer_last_timestep = (
                f"{self.get_array_name(self.variables['_last_timestep'])}[_idx]"
            )
        else:
            pointer_last_timestep = None

        if options["save_failed_steps"]:
            N = self.variables["N"].item()
            try:
                self.owner.variables.add_array("_failed_steps", size=N, dtype=np.int32)
            except KeyError:
                # has already been run
                pass
            self.variables["_failed_steps"] = self.owner.variables.get("_failed_steps")
            pointer_failed_steps = (
                f"{self.get_array_name(self.variables['_failed_steps'])}[_idx]"
            )
        else:
            pointer_failed_steps = None

        if options["save_step_count"]:
            N = int(self.variables["N"].get_value())
            try:
                self.owner.variables.add_array("_step_count", size=N, dtype=np.int32)
            except KeyError:
                # has already been run
                pass
            self.variables["_step_count"] = self.owner.variables.get("_step_count")
            pointer_step_count = (
                f"{self.get_array_name(self.variables['_step_count'])}[_idx]"
            )
        else:
            pointer_step_count = None

        return {
            "pointer_last_timestep": pointer_last_timestep,
            "pointer_failed_steps": pointer_failed_steps,
            "pointer_step_count": pointer_step_count,
        }

    def translate(
        self, code, dtype
    ):  # TODO: it's not so nice we have to copy the contents of this function..
        """
        Translates an abstract code block into the target language.
        """
        # first check if user code is not using variables that are also used by GSL
        reserved_variables = [
            "_dataholder",
            "_fill_y_vector",
            "_empty_y_vector",
            "_GSL_dataholder",
            "_GSL_y",
            "_GSL_func",
        ]
        if any([var in self.variables for var in reserved_variables]):
            # import here to avoid circular import
            raise ValueError(
                f"The variables {str(reserved_variables)} are reserved for the GSL"
                " internal code."
            )

        # if the following statements are not added, Brian translates the
        # differential expressions in the abstract code for GSL to scalar statements
        # in the case no non-scalar variables are used in the expression
        diff_vars = self.find_differential_variables(list(code.values()))
        self.add_gsl_variables_as_non_scalar(diff_vars)

        # add arrays we want to use in generated code before self.generator.translate() so
        # brian does namespace unpacking for us
        pointer_names = self.add_meta_variables(self.method_options)

        scalar_statements = {}
        vector_statements = {}
        for ac_name, ac_code in code.items():
            statements = make_statements(
                ac_code, self.variables, dtype, optimise=True, blockname=ac_name
            )
            scalar_statements[ac_name], vector_statements[ac_name] = statements
        for vs in vector_statements.values():
            # Check that the statements are meaningful independent on the order of
            # execution (e.g. for synapses)
            try:
                if self.has_repeated_indices(
                    vs
                ):  # only do order dependence if there are repeated indices
                    check_for_order_independence(
                        vs, self.generator.variables, self.generator.variable_indices
                    )
            except OrderDependenceError:
                # If the abstract code is only one line, display it in full
                if len(vs) <= 1:
                    error_msg = f"Abstract code: '{vs[0]}'\n"
                else:
                    error_msg = (
                        f"{len(vs)} lines of abstract code, first line is: '{vs[0]}'\n"
                    )
                logger.warn(
                    "Came across an abstract code block that may not be "
                    "well-defined: the outcome may depend on the "
                    "order of execution. You can ignore this warning if "
                    "you are sure that the order of operations does not "
                    "matter. " + error_msg
                )

        # save function names because self.generator.translate_statement_sequence
        # deletes these from self.variables but we need to know which identifiers
        # we can safely ignore (i.e. we can ignore the functions because they are
        # handled by the original generator)
        self.function_names = self.find_function_names()

        scalar_code, vector_code, kwds = self.generator.translate_statement_sequence(
            scalar_statements, vector_statements
        )

        ############ translate code for GSL

        # first check if any indexing other than '_idx' is used (currently not supported)
        for code_list in list(scalar_code.values()) + list(vector_code.values()):
            for code in code_list:
                m = re.search(r"\[(\w+)\]", code)
                if m is not None:
                    if m.group(1) != "0" and m.group(1) != "_idx":
                        from brian2.stateupdaters.base import (
                            UnsupportedEquationsException,
                        )

                        raise UnsupportedEquationsException(
                            "Equations result in state "
                            "updater code with indexing "
                            "other than '_idx', which "
                            "is currently not supported "
                            "in combination with the "
                            "GSL stateupdater."
                        )

        # differential variable specific operations
        to_replace = self.diff_var_to_replace(diff_vars)
        GSL_support_code = self.get_dimension_code(len(diff_vars))
        GSL_support_code += self.yvector_code(diff_vars)

        # analyze all needed variables; if not in self.variables: put in separate dic.
        # also keep track of variables needed for scalar statements and vector statements
        other_variables = self.find_undefined_variables(
            scalar_statements[None] + vector_statements[None]
        )
        variables_in_scalar = self.find_used_variables(
            scalar_statements[None], other_variables
        )
        variables_in_vector = self.find_used_variables(
            vector_statements[None], other_variables
        )
        # so that _dataholder holds diff_vars as well, even if they don't occur
        # in the actual statements
        for var in list(diff_vars.keys()):
            if var not in variables_in_vector:
                variables_in_vector[var] = self.variables[var]
        # let's keep track of the variables that eventually need to be added to
        # the _GSL_dataholder somehow
        self.variables_to_be_processed = list(variables_in_vector.keys())

        # add code for _dataholder struct
        GSL_support_code = self.write_dataholder(variables_in_vector) + GSL_support_code
        # add e.g. _lio_1 --> _GSL_dataholder._lio_1 to replacer
        to_replace.update(
            self.to_replace_vector_vars(
                variables_in_vector, ignore=list(diff_vars.keys())
            )
        )
        # write statements that unpack (python) namespace to _dataholder struct
        # or local namespace
        GSL_main_code = self.unpack_namespace(
            variables_in_vector, variables_in_scalar, ["t"]
        )

        # rewrite actual calculations described by vector_code and put them in _GSL_func
        func_code = self.translate_one_statement_sequence(
            vector_statements[None], scalar=False
        )
        GSL_support_code += self.make_function_code(
            self.translate_vector_code(func_code, to_replace)
        )
        scalar_func_code = self.translate_one_statement_sequence(
            scalar_statements[None], scalar=True
        )
        # rewrite scalar code, keep variables that are needed in scalar code normal
        # and add variables to _dataholder for vector_code
        GSL_main_code += f"\n{self.translate_scalar_code(scalar_func_code, variables_in_scalar, variables_in_vector)}"
        if len(self.variables_to_be_processed) > 0:
            raise AssertionError(
                "Not all variables that will be used in the vector "
                "code have been added to the _GSL_dataholder. This "
                "might mean that the _GSL_func is using "
                "uninitialized variables.\n"
                "The unprocessed variables "
                f"are: {self.variables_to_be_processed}"
            )

        scalar_code["GSL"] = GSL_main_code
        kwds["define_GSL_scale_array"] = self.scale_array_code(
            diff_vars, self.method_options
        )
        kwds["n_diff_vars"] = len(diff_vars)
        kwds["GSL_settings"] = dict(self.method_options)
        kwds["GSL_settings"]["integrator"] = self.integrator
        kwds["support_code_lines"] += GSL_support_code.split("\n")
        kwds["t_array"] = f"{self.get_array_name(self.variables['t'])}[0]"
        kwds["dt_array"] = f"{self.get_array_name(self.variables['dt'])}[0]"
        kwds["define_dt"] = "dt" not in variables_in_scalar
        kwds["cpp_standalone"] = self.is_cpp_standalone()
        for key, value in list(pointer_names.items()):
            kwds[key] = value
        return scalar_code, vector_code, kwds


class GSLCythonCodeGenerator(GSLCodeGenerator):
    syntax = {
        "end_statement": "",
        "access_pointer": ".",
        "start_declare": "cdef extern ",
        "open_function": ":",
        "open_struct": ":",
        "end_function": "",
        "end_struct": "",
        "open_cast": "<",
        "close_cast": ">",
        "diff_var_declaration": "",
    }

    def c_data_type(self, dtype):
        return c_data_type(dtype)

    def initialize_array(self, varname, values):
        value_list = ", ".join(repr(v) for v in values)
        code = "cdef double {varname}[{n_values}]\n"
        code += "{varname}[:] = [{value_list}]"
        return code.format(varname=varname, value_list=value_list, n_values=len(values))

    def var_replace_diff_var_lhs(self, var, ind):
        return {f"_gsl_{var}_f{ind}": f"f[{ind}]"}

    def var_init_lhs(self, var, type):
        return var

    def unpack_namespace_single(self, var_obj, in_vector, in_scalar):
        code = []
        if isinstance(var_obj, ArrayVariable):
            array_name = self.generator.get_array_name(var_obj)
            dtype = self.c_data_type(var_obj.dtype)
            if in_vector:
                code += [
                    f"_GSL_dataholder.{array_name} = <{dtype} *> _buf_{array_name}.data"
                ]
            if in_scalar:
                code += [f"{array_name} = <{dtype} *> _buf_{array_name}.data"]
        else:
            if in_vector:
                code += [
                    f'_GSL_dataholder.{var_obj.name} = _namespace["{var_obj.name}"]'
                ]
            if in_scalar:
                code += [f'{var_obj.name} = _namespace["{var_obj.name}"]']
        return "\n".join(code)

    @staticmethod
    def get_array_name(var, access_data=True):
        # We have to do the import here to avoid circular import dependencies.
        from brian2.codegen.generators.cython_generator import CythonCodeGenerator

        return CythonCodeGenerator.get_array_name(var, access_data)


class GSLCPPCodeGenerator(GSLCodeGenerator):
    def __getattr__(self, item):
        return getattr(self.generator, item)

    syntax = {
        "end_statement": ";",
        "access_pointer": "->",
        "start_declare": 'extern "C" ',
        "open_function": "\n{",
        "open_struct": "\n{",
        "end_function": "\n}",
        "end_struct": "\n};",
        "open_cast": "(",
        "close_cast": ")",
        "diff_var_declaration": "const scalar ",
    }

    def c_data_type(self, dtype):
        return self.generator.c_data_type(dtype)

    def initialize_array(self, varname, values):
        value_list = ", ".join(repr(v) for v in values)
        return f"double const {varname}[] = {{{value_list}}};"

    def var_replace_diff_var_lhs(self, var, ind):
        scalar_dtype = self.c_data_type(prefs.core.default_float_dtype)
        f = f"f[{ind}]"
        try:
            if "unless refractory" in self.variable_flags[var]:
                return {
                    f"_gsl_{var}_f{ind}": f,
                    f"{scalar_dtype} _gsl_{var}_f{ind};": "",
                    f"{scalar_dtype} {f};": "",
                }  # in case the replacement of
                # _gsl_var_find to f[ind] happens
                # first
        except KeyError:
            pass
        return {f"const {scalar_dtype} _gsl_{var}_f{ind}": f}

    def var_init_lhs(self, var, type):
        return type + var

    def unpack_namespace_single(self, var_obj, in_vector, in_scalar):
        if isinstance(var_obj, ArrayVariable):
            pointer_name = self.get_array_name(var_obj, access_data=True)
            array_name = self.get_array_name(var_obj)
            if in_vector:
                return f"_GSL_dataholder.{pointer_name} = {array_name};"
            else:
                return ""
        else:
            if in_vector:
                return f"_GSL_dataholder.{var_obj.name} = {var_obj.name};"
            else:
                return ""



"""
Brette R (2013). Sharpness of spike initiation in neurons explained by compartmentalization.
PLoS Comp Biol, doi: 10.1371/journal.pcbi.1003338.

Fig. 3. A, B. Kink with only Nav1.6 channels
"""
from brian2 import *
from params import *

codegen.target='numpy'

defaultclock.dt = 0.025*ms

# Morphology
morpho = Soma(50*um)  # chosen for a target Rm
morpho.axon = Cylinder(diameter=1*um, length=300*um, n=300)

location = 40*um  # where Na channels are placed

# Channels
eqs='''
Im = gL*(EL - v) + gNa*m*(ENa - v) : amp/meter**2
dm/dt = (minf - m) / taum : 1 # simplified Na channel
minf = 1 / (1 + exp((va - v) / ka)) : 1
gNa : siemens/meter**2
Iin : amp (point current)
'''

neuron = SpatialNeuron(morphology=morpho, model=eqs, Cm=Cm, Ri=Ri,
                       method="exponential_euler")

compartment = morpho.axon[location]
neuron.v = EL
neuron.gNa[compartment] = gNa_0/neuron.area[compartment]
M = StateMonitor(neuron, ['v', 'm'], record=True)

run(20*ms, report='text')
neuron.Iin[0] = gL * 20*mV * neuron.area[0]
run(80*ms, report='text')

subplot(121)
plot(M.t/ms, M[0].v/mV, 'r')
plot(M.t/ms, M[compartment].v/mV, 'k')
plot(M.t/ms, M[compartment].m*(80+60)-80, 'k--')  # open channels
ylim(-80, 60)
xlabel('Time (ms)')
ylabel('V (mV)')
title('Voltage traces')

subplot(122)
dm = diff(M[0].v) / defaultclock.dt
dm40 = diff(M[compartment].v) / defaultclock.dt
plot((M[0].v/mV)[1:], dm/(volt/second), 'r')
plot((M[compartment].v/mV)[1:], dm40/(volt/second), 'k')
xlim(-80, 40)
xlabel('V (mV)')
ylabel('dV/dt (V/s)')
title('Phase plot')
tight_layout()
show()



#!/usr/bin/env python3
"""
Fig. 1 from:

M. Tsodyks, K. Pawelzik, H. Markram
Neural Networks with Dynamic Synapses
Neural Computation 10, 821–835 (1998)

https://doi.org/10.1162/089976698300017502

Sebastian Schmitt, 2022
"""

import numpy as np
import matplotlib.pyplot as plt

from brian2 import (
    NeuronGroup,
    Synapses,
    SpikeGeneratorGroup,
    SpikeMonitor,
    StateMonitor,
)
from brian2 import ms, mV, pA, Mohm, Gohm, Hz
from brian2 import run


def get_neuron(tau_mem, R_in):
    """
    tau_mem -- membrane time constant
    R_in -- input resistance
    """
    neuron = NeuronGroup(1,
                         """
                         tau_mem : second
                         I_syn : ampere
                         R_in : ohm
                         dv/dt = -v/tau_mem + (R_in*I_syn)/tau_mem : volt
                         """,
                         method="exact")

    neuron.tau_mem = tau_mem
    neuron.R_in = R_in

    return neuron


def get_synapses(stimulus, neuron, tau_inact, A_SE, U_SE, tau_rec, tau_facil=None):
    """
    stimulus -- input stimulus
    neuron -- target neuron
    tau_inact -- inactivation time constant
    A_SE -- absolute synaptic strength
    U_SE -- utilization of synaptic efficacy
    tau_rec -- recovery time constant
    tau_facil -- facilitation time constant (optional)
    """

    synapses_eqs = """
    dx/dt =  z/tau_rec   : 1 (clock-driven) # recovered
    dy/dt = -y/tau_inact : 1 (clock-driven) # active
    A_SE : ampere
    U_SE : 1
    tau_inact : second
    tau_rec : second
    z = 1 - x - y : 1 # inactive
    I_syn_post = A_SE*y : ampere (summed)
    """

    if tau_facil:
        synapses_eqs += """
        du/dt = -u/tau_facil : 1 (clock-driven)
        tau_facil : second
        """

        synapses_action = """
        u += U_SE*(1-u)
        y += u*x # important: update y first
        x += -u*x
        """
    else:
        synapses_action = """
        y += U_SE*x # important: update y first
        x += -U_SE*x
        """

    synapses = Synapses(stimulus,
                        neuron,
                        model=synapses_eqs,
                        on_pre=synapses_action,
                        method="exponential_euler")
    synapses.connect()

    # start fully recovered
    synapses.x = 1

    synapses.tau_inact = tau_inact
    synapses.A_SE = A_SE
    synapses.U_SE = U_SE
    synapses.tau_rec = tau_rec

    if tau_facil:
        synapses.tau_facil = tau_facil

    return synapses


def get_stimulus(start, stop, frequency):
    """
    start -- start time of stimulus
    stop -- stop time of stimulus
    frequency -- frequency of stimulus
    """

    times = np.arange(start / ms, stop / ms, 1 / (frequency / Hz) * 1e3) * ms
    stimulus = SpikeGeneratorGroup(1, [0] * len(times), times)

    return stimulus


parameters = {
    "A": {
        "neuron": {"tau_mem": 40 * ms,
                   "R_in": 100*Mohm},
        "synapse": {
            "tau_inact": 3 * ms,
            "A_SE": 250 * pA,
            "tau_rec": 800 * ms,
            "U_SE": 0.6, # 0.5 from publication does not match plot
        },
        "stimulus": {"start": 100 * ms,
                     "stop": 1100 * ms,
                     "frequency": 20 * Hz},
        "simulation": {"duration": 1200 * ms},
        "plot": {
            "title": "A) D - 20 Hz",
            "ylim": [0, 1],
            "xlim": [0, 1200],
            "xtickstep": 200,
        },
    },
    "B": {
        "neuron": {"tau_mem": 60 * ms,
                   "R_in": 1*Gohm},
        "synapse": {
            "tau_inact": 1.5 * ms,
            "A_SE": 1540 * pA,
            "tau_rec": 130 * ms,
            "U_SE": 0.03,
            "tau_facil": 530 * ms,
        },
        "stimulus": {"start": 100 * ms,
                     "stop": 1100 * ms,
                     "frequency": 20 * Hz},
        "simulation": {"duration": 1200 * ms},
        "plot": {
            "title": "B) F - 20 Hz",
            "ylim": [0, 14.9],
            "xlim": [0, 1200],
            "xtickstep": 200,
        },
    },
    "C": {
        "neuron": {"tau_mem": 60 * ms,
                   "R_in": 1*Gohm},
        "synapse": {
            "tau_inact": 1.5 * ms,
            "A_SE": 1540 * pA,
            "tau_rec": 130 * ms,
            "U_SE": 0.03,
            "tau_facil": 530 * ms,
        },
        "stimulus": {"start": 100 * ms,
                     "stop": 375 * ms,
                     "frequency": 70 * Hz},
        "simulation": {"duration": 500 * ms},
        "plot": {
            "title": "C) F - 70 Hz",
            "ylim": [0, 20],
            "xlim": [0, 500],
            "xtickstep": 50,
        },
    },
}

fig, axes = plt.subplots(3)

for ax, (panel, p) in zip(axes, parameters.items()):

    neuron = get_neuron(**p["neuron"])
    stimulus = get_stimulus(**p["stimulus"])
    synapses = get_synapses(stimulus, neuron, **p["synapse"])

    state_monitor_neuron = StateMonitor(neuron, ["v"], record=True)

    run(p["simulation"]["duration"])

    ax.plot(
        state_monitor_neuron.t / ms,
        state_monitor_neuron[0].v / mV,
        label=p["plot"]["title"],
    )

    ax.set_xlim(*p["plot"]["xlim"])
    ax.set_ylim(*p["plot"]["ylim"])
    ax.set_ylabel("mV")
    ax.set_xlabel("Time (ms)")

    ax.set_xticks(
        np.arange(
            p["plot"]["xlim"][0],
            p["plot"]["xlim"][1] + p["plot"]["xtickstep"],
            p["plot"]["xtickstep"],
        )
    )

    ax.legend()

plt.show()



"""
Hodgkin-Huxley equations (1952)

We calculate the extracellular field potential at various places.
"""
from brian2 import *
defaultclock.dt = 0.01*ms
morpho = Cylinder(x=[0, 10]*cm, diameter=2*238*um, n=1000, type='axon')

El = 10.613* mV
ENa = 115*mV
EK = -12*mV
gl = 0.3*msiemens/cm**2
gNa0 = 120*msiemens/cm**2
gK = 36*msiemens/cm**2

# Typical equations
eqs = '''
# The same equations for the whole neuron, but possibly different parameter values
# distributed transmembrane current
Im = gl * (El-v) + gNa * m**3 * h * (ENa-v) + gK * n**4 * (EK-v) : amp/meter**2
I : amp (point current) # applied current
dm/dt = alpham * (1-m) - betam * m : 1
dn/dt = alphan * (1-n) - betan * n : 1
dh/dt = alphah * (1-h) - betah * h : 1
alpham = (0.1/mV) * 10*mV/exprel((-v+25*mV)/(10*mV))/ms : Hz
betam = 4 * exp(-v/(18*mV))/ms : Hz
alphah = 0.07 * exp(-v/(20*mV))/ms : Hz
betah = 1/(exp((-v+30*mV) / (10*mV)) + 1)/ms : Hz
alphan = (0.01/mV) * 10*mV/exprel((-v+10*mV)/(10*mV))/ms : Hz
betan = 0.125*exp(-v/(80*mV))/ms : Hz
gNa : siemens/meter**2
'''

neuron = SpatialNeuron(morphology=morpho, model=eqs, Cm=1*uF/cm**2,
                       Ri=35.4*ohm*cm, method="exponential_euler")
neuron.v = 0*mV
neuron.h = 1
neuron.m = 0
neuron.n = .5
neuron.I = 0
neuron.gNa = gNa0
neuron[5*cm:10*cm].gNa = 0*siemens/cm**2
M = StateMonitor(neuron, 'v', record=True)

# LFP recorder
Ne = 5 # Number of electrodes
sigma = 0.3*siemens/meter # Resistivity of extracellular field (0.3-0.4 S/m)
lfp = NeuronGroup(Ne, model='''v : volt
                               x : meter
                               y : meter
                               z : meter''')
lfp.x = 7*cm # Off center (to be far from stimulating electrode)
lfp.y = [1*mm, 2*mm, 4*mm, 8*mm, 16*mm]
S = Synapses(neuron, lfp, model='''w : ohm*meter**2 (constant) # Weight in the LFP calculation
                                   v_post = w*(Ic_pre-Im_pre) : volt (summed)''')
S.summed_updaters['v_post'].when = 'after_groups'  # otherwise Ic has not yet been updated for the current time step.
S.connect()
S.w = 'area_pre/(4*pi*sigma)/((x_pre-x_post)**2+(y_pre-y_post)**2+(z_pre-z_post)**2)**.5'

Mlfp = StateMonitor(lfp, 'v', record=True)

run(50*ms, report='text')
neuron.I[0] = 1*uA  # current injection at one end
run(3*ms)
neuron.I = 0*amp
run(100*ms, report='text')

subplot(211)
for i in range(10):
    plot(M.t/ms, M.v[i*100]/mV)
ylabel('$V_m$ (mV)')
subplot(212)
for i in range(5):
    plot(M.t/ms, Mlfp.v[i]/mV)
ylabel('LFP (mV)')
xlabel('Time (ms)')
show()



#!/usr/bin/env python3
"""
Ornstein-Uhlenbeck process

Figure 2: Two realizations of the Ornstein-Uhlenbeck process for
parameters τ=1.0 and σ=0.1 (black curve), and for τ=0.1 and σ=0.31622
(red curve). In both cases the noise intensity is σ^2*τ=0.01 . The red
curve represents a noise that more closely mimics Gaussian white
noise. Both realizations begin here at x(0)=1.0 , after which the mean
decays exponentially to zero with time constant τ.

Andre Longtin (2010) Stochastic dynamical systems. Scholarpedia, 5(4):1619.

Sebastian Schmitt, 2022
"""

import matplotlib.pyplot as plt
import numpy as np

from brian2 import run
from brian2 import NeuronGroup, StateMonitor
from brian2 import second, ms

N = NeuronGroup(
    2,
    """
    tau : second
    sigma : 1
    dy/dt = -y/tau + sqrt(2*sigma**2/tau)*xi : 1
    """,
    method="euler"
)

N.tau = np.array([1, 0.1]) * second
N.sigma = np.array([0.1, 0.31622])
N.y = 1

M = StateMonitor(N, "y", record=True)

run(10 * second)

plt.plot(M.t / second, M.y[1], color="red", label=r"$\tau$=0.1 s, $\sigma$=0.31622")
plt.plot(M.t / second, M.y[0], color="k", label=r"$\tau$=1 s, $\sigma$=0.1")

plt.xlim(0, 10)
plt.ylim(-1.1, 1.1)

plt.xlabel("time (sec)")
plt.ylabel("Ornstein-Uhlenbeck process")

plt.legend()

plt.show()



"""
Demonstrate the usage of the `Morphology` object.
"""
from brian2 import *

# Morphology
morpho = Soma(30*um)
morpho.L = Cylinder(diameter=1*um, length=100*um, n=5)
morpho.LL = Cylinder(diameter=1*um, length=20*um, n=2)
morpho.R = Cylinder(diameter=1*um, length=100*um, n=5)

# Passive channels
gL = 1e-4*siemens/cm**2
EL = -70*mV
eqs = '''
Im = gL * (EL-v) : amp/meter**2
'''

neuron = SpatialNeuron(morphology=morpho, model=eqs,
                       Cm=1*uF/cm**2, Ri=100*ohm*cm, method='exponential_euler')
neuron.v = arange(0, 13)*volt

print(neuron.v)
print(neuron.L.v)
print(neuron.LL.v)
print(neuron.L.main.v)



"""
Synaptic model with continuous interaction
------------------------------------------

This example implements a conductance base synapse that is continuously linking
two neurons, i.e. the synaptic gating variable updates at each time step.
Two Reduced Traub-Miles Model (RTM) neurons are connected to each other
through a directed synapse from neuron 1 to 2.

Here, the complexity stems from the fact that the synaptic conductance is a
continuous function of the membrane potential, instead of being triggered by
individual spikes. This can be useful in particular when analyzing models
mathematically but it is not recommended in most cases because they tend to be
less efficient. Also note that this model only works with (pre-synaptic) neuron
models that model the action potential in detail, i.e. not with
integrate-and-fire type models.

There are two broad approaches (``s`` as part of the pre-synaptic neuron or
``s`` as part of the Synapses object), all depends on whether the time
constants are the same across all synapses or whether they can vary between
synapses. In this example, the time constant is assumed to be the same and
``s`` is therefore part of the pre-synaptic neuron model.

References:

-  Introduction to modeling neural dynamics, Börgers, chapter 20
- `Discussion in Brian forum <https://brian.discourse.group/t/how-to-implement-a-conductance-base-synapse/77/2>`_
"""
from brian2 import *

I_e = 1.5*uA
simulation_time = 100*ms
# neuron RTM parameters
El = -67 * mV
EK = -100 * mV
ENa = 50 * mV
ESyn = 0 * mV
gl = 0.1 * msiemens
gK = 80 * msiemens
gNa = 100 * msiemens

C = 1 * ufarad

weight = 0.25
gSyn = 1.0 * msiemens
tau_d = 2 * ms
tau_r = 0.2 * ms

# forming RTM model with differential equations
eqs = """
alphah = 0.128 * exp(-(vm + 50.0*mV) / (18.0*mV))/ms :Hz
alpham = 0.32/mV * (vm + 54*mV) / (1.0 - exp(-(vm + 54.0*mV) / (4.0*mV)))/ms:Hz
alphan = 0.032/mV * (vm + 52*mV) / (1.0 - exp(-(vm + 52.0*mV) / (5.0*mV)))/ms:Hz

betah  = 4.0 / (1.0 + exp(-(vm + 27.0*mV) / (5.0*mV)))/ms:Hz
betam  = 0.28/mV * (vm + 27.0*mV) / (exp((vm + 27.0*mV) / (5.0*mV)) - 1.0)/ms:Hz
betan  = 0.5 * exp(-(vm + 57.0*mV) / (40.0*mV))/ms:Hz

membrane_Im = I_ext + gNa*m**3*h*(ENa-vm) +
              gl*(El-vm) + gK*n**4*(EK-vm) + gSyn*s_in*(-vm): amp
I_ext : amp
s_in  : 1

dm/dt = alpham*(1-m)-betam*m : 1
dn/dt = alphan*(1-n)-betan*n : 1
dh/dt = alphah*(1-h)-betah*h : 1

ds/dt = 0.5 * (1 + tanh(0.1*vm/mV)) * (1-s)/tau_r - s/tau_d : 1

dvm/dt = membrane_Im/C : volt
"""

neuron = NeuronGroup(2, eqs, method="exponential_euler")

# initialize variables
neuron.vm = [-70.0, -65.0]*mV
neuron.m = "alpham / (alpham + betam)"
neuron.h = "alphah / (alphah + betah)"
neuron.n = "alphan / (alphan + betan)"
neuron.I_ext = [I_e, 0.0*uA]

S = Synapses(neuron,
             neuron,
             's_in_post = weight*s_pre:1 (summed)')
S.connect(i=0, j=1)

# tracking variables
st_mon = StateMonitor(neuron, ["vm", "s", "s_in"], record=[0, 1])

# running the simulation
run(simulation_time)

# plot the results
fig, ax = plt.subplots(2, figsize=(10, 6), sharex=True,
                       gridspec_kw={'height_ratios': (3, 1)})

ax[0].plot(st_mon.t/ms, st_mon.vm[0]/mV,
           lw=2, c="r", alpha=0.5, label="neuron 0")
ax[0].plot(st_mon.t/ms, st_mon.vm[1]/mV,
           lw=2, c="b", alpha=0.5, label='neuron 1')
ax[1].plot(st_mon.t/ms, st_mon.s[0],
           lw=2, c="r", alpha=0.5, label='s, neuron 0')
ax[1].plot(st_mon.t/ms, st_mon.s_in[1],
           lw=2, c="b", alpha=0.5, label='s_in, neuron 1')
ax[0].set(ylabel='v [mV]', xlim=(0, np.max(st_mon.t / ms)),
          ylim=(-100, 50))
ax[1].set(xlabel="t [ms]", ylabel="s", ylim=(0, 1))

ax[0].legend()
ax[1].legend()

plt.show()



"""
Example demonstrating the use of custom events.

Here we have three neurons, the first is Poisson spiking and connects to neuron G,
which in turn connects to neuron H. Neuron G has two variables v and g, and the
incoming Poisson spikes cause an instantaneous increase in variable g. g decays
rapidly, and in turn causes a slow increase in v. If v crosses a threshold, it
causes a standard spike and reset. If g crosses a threshold, it causes a custom
event ``gspike``, and if it returns below that threshold it causes a custom
event ``end_gspike``. The standard spike event when v crosses a threshold
causes an instantaneous increase in variable x in neuron H (which happens
through the standard ``pre`` pathway in the synapses), and the gspike
event causes an increase in variable y (which happens through the custom
pathway ``gpath``).
"""
from brian2 import *
# Input Poisson spikes
inp = PoissonGroup(1, rates=250*Hz)
# First group G
eqs_G = '''
dv/dt = (g-v)/(50*ms) : 1
dg/dt = -g/(10*ms) : 1
allow_gspike : boolean
'''
G = NeuronGroup(1, eqs_G, threshold='v>1',
                reset='v = 0; g = 0; allow_gspike = True;',
                events={'gspike': 'g>1 and allow_gspike',
                        'end_gspike': 'g<1 and not allow_gspike'})
G.run_on_event('gspike', 'allow_gspike = False')
G.run_on_event('end_gspike', 'allow_gspike = True')
# Second group H
eqs_H = '''
dx/dt = -x/(10*ms) : 1
dy/dt = -y/(10*ms) : 1
'''
H = NeuronGroup(1, eqs_H)
# Synapses from input Poisson group to G
Sin = Synapses(inp, G, on_pre='g += 0.5')
Sin.connect()
# Synapses from G to H
S = Synapses(G, H,
             on_pre={'pre': 'x += 1',
                     'gpath': 'y += 1'},
             on_event={'pre': 'spike',
                       'gpath': 'gspike'})
S.connect()
# Monitors
Mstate = StateMonitor(G, ('v', 'g'), record=True)
Mgspike = EventMonitor(G, 'gspike', 'g')
Mspike = SpikeMonitor(G, 'v')
MHstate = StateMonitor(H, ('x', 'y'), record=True)
# Initialise and run
G.allow_gspike = True
run(500*ms)
# Plot
figure(figsize=(10, 4))
subplot(121)
plot(Mstate.t/ms, Mstate.g[0], '-g', label='g')
plot(Mstate.t/ms, Mstate.v[0], '-b', lw=2, label='V')
plot(Mspike.t/ms, Mspike.v, 'ob', label='_nolegend_')
plot(Mgspike.t/ms, Mgspike.g, 'og', label='_nolegend_')
xlabel('Time (ms)')
title('Presynaptic group G')
legend(loc='best')
subplot(122)
plot(MHstate.t/ms, MHstate.y[0], '-r', label='y')
plot(MHstate.t/ms, MHstate.x[0], '-k', lw=2, label='x')
xlabel('Time (ms)')
title('Postsynaptic group H')
legend(loc='best')
tight_layout()
show()


import sys, os, numpy, time, pylab

filename = 'CUBA_standalone.py'
datapath = 'data_cuba'
threads  = [0, 1, 2, 4, 6]
results  = {}
results['duration'] = []

for t in threads:
    start = time.time()
    os.system('python %s 1 %d' %(filename, t))
    with open('%s_%d/speed.txt' %(datapath, t), 'r') as f:
        results['duration'] += [float(f.read())]


for t in threads:
    results[t] = {}
    path       = datapath + '_%d/' %t
    ids        = numpy.fromfile(path+'results/_dynamic_array_spikemonitor_i', dtype=numpy.int32)
    times      = numpy.fromfile(path+'results/_dynamic_array_spikemonitor_t', dtype=numpy.float64)
    vms        = numpy.fromfile(path+'results/_array_neurongroup_v', dtype=numpy.float64)
    w          = numpy.fromfile(path+'results/_dynamic_array_synapses_1_w', dtype=numpy.float64)
    results[t]['spikes'] = (times, ids)
    results[t]['vms']    = vms
    results[t]['w']      = w

pylab.figure()
pylab.subplot(221)
pylab.title('Raster plots')
pylab.xlabel('Time [s]')
pylab.ylabel('# cell')
for t in threads:
    pylab.plot(results[t]['spikes'][0], results[t]['spikes'][1], '.')
pylab.legend([str(t) for t in threads])
pylab.xlim(0.5, 0.55)

pylab.subplot(222)
pylab.title('Mean Rates')
pylab.xlabel('Time [s]')
pylab.ylabel('Rate [Hz]')
for t in threads:
    x,  y = numpy.histogram(results[t]['spikes'][0], 100)
    pylab.plot(y[1:], x)
pylab.legend([str(t) for t in threads])


pylab.subplot(223)
pylab.title('Network')
pylab.xlabel('# threads')
pylab.ylabel('Number of synapses')
r = []
for t in threads:
    r += [len(results[t]['w'])]
pylab.bar(threads, r)
pylab.ylim(min(r)-1000, max(r)+1000)

pylab.subplot(224)
pylab.title('Speed')
pylab.plot(threads, results['duration'])
pylab.xlabel('# threads')
pylab.ylabel('Time [s]')

pylab.tight_layout()
pylab.savefig('cuba_openmp.png')


pylab.show()




import multiprocessing
import os

import pytest

from brian2.core.preferences import prefs
from brian2.utils.logger import BrianLogger, catch_logs, get_logger

logger = get_logger("brian2.tests.test_logger")


@pytest.mark.codegen_independent
def test_file_logging():
    BrianLogger.initialize()
    logger.error("error message xxx")
    logger.warn("warning message xxx")
    logger.info("info message xxx")
    logger.debug("debug message xxx")
    logger.diagnostic("diagnostic message xxx")
    BrianLogger.file_handler.flush()
    # By default, only >= debug messages should show up
    assert os.path.isfile(BrianLogger.tmp_log)
    with open(BrianLogger.tmp_log, encoding="utf-8") as f:
        log_content = f.readlines()
    for level, line in zip(["error", "warning", "info", "debug"], log_content[-4:]):
        assert "brian2.tests.test_logger" in line
        assert f"{level} message xxx" in line
        assert level.upper() in line


@pytest.mark.codegen_independent
def test_file_logging_special_characters():
    BrianLogger.initialize()
    # Test logging with special characters that could occur in log messages and
    # require UTF-8
    special_chars = "→ ≠ ≤ ≥ ← ∞ µ ∝ ∂ ∅"
    logger.debug(special_chars)
    BrianLogger.file_handler.flush()
    assert os.path.isfile(BrianLogger.tmp_log)
    with open(BrianLogger.tmp_log, encoding="utf-8") as f:
        log_content = f.readlines()
    last_line = log_content[-1]
    assert "brian2.tests.test_logger" in last_line
    assert special_chars in last_line


def run_in_process(x):
    logger.info(f"subprocess info message {x}")


def run_in_process_with_logger(x):
    prefs.logging.delete_log_on_exit = False
    BrianLogger.initialize()
    logger.info(f"subprocess info message {x}")
    BrianLogger.file_handler.flush()
    return BrianLogger.tmp_log


@pytest.mark.codegen_independent
def test_file_logging_multiprocessing():
    logger.info("info message before multiprocessing")
    p = multiprocessing.Pool()

    try:
        p.map(run_in_process, range(3))
    finally:
        p.close()
        p.join()

    BrianLogger.file_handler.flush()
    assert os.path.isfile(BrianLogger.tmp_log)
    with open(BrianLogger.tmp_log, encoding="utf-8") as f:
        log_content = f.readlines()
    # The subprocesses should not have written to the log file
    assert "info message before multiprocessing" in log_content[-1]


@pytest.mark.codegen_independent
def test_file_logging_multiprocessing_with_loggers():
    logger.info("info message before multiprocessing")

    p = multiprocessing.Pool()
    try:
        log_files = p.map(run_in_process_with_logger, range(3))
    finally:
        p.close()
        p.join()

    BrianLogger.file_handler.flush()
    assert os.path.isfile(BrianLogger.tmp_log)
    with open(BrianLogger.tmp_log, encoding="utf-8") as f:
        log_content = f.readlines()
    # The subprocesses should not have written to the main log file
    assert "info message before multiprocessing" in log_content[-1]

    # Each subprocess should have their own log file
    for x, log_file in enumerate(log_files):
        assert os.path.isfile(log_file)
        with open(log_file, encoding="utf-8") as f:
            log_content = f.readlines()
        assert f"subprocess info message {x}" in log_content[-1]

    prefs.logging.delete_log_on_exit = True


@pytest.mark.codegen_independent
def test_submodule_logging():
    submodule_logger = get_logger("submodule.dummy")
    BrianLogger.initialize()
    submodule_logger.error("error message xxx")
    submodule_logger.warn("warning message xxx")
    submodule_logger.info("info message xxx")
    submodule_logger.debug("debug message xxx")
    submodule_logger.diagnostic("diagnostic message xxx")
    BrianLogger.file_handler.flush()
    # By default, only >= debug messages should show up
    assert os.path.isfile(BrianLogger.tmp_log)
    with open(BrianLogger.tmp_log, encoding="utf-8") as f:
        log_content = f.readlines()
    for level, line in zip(["error", "warning", "info", "debug"], log_content[-4:]):
        assert "submodule.dummy" in line
        # The logger name has brian2 internally prefixed, but this shouldn't show up in logs
        assert not "brian2.submodule.dummy" in line
        assert f"{level} message xxx" in line
        assert level.upper() in line

    with catch_logs() as l:
        logger.warn("warning message from Brian")
        submodule_logger.warn("warning message from submodule")
    # only the warning from Brian should be logged
    assert len(l) == 1
    assert "warning message from Brian" in l[0]

    with catch_logs(only_from=("submodule",)) as l:
        logger.warn("warning message from Brian")
        submodule_logger.warn("warning message from submodule")
    # only the warning from submodule should be logged
    assert len(l) == 1
    assert "warning message from submodule" in l[0]

    # Make sure that a submodule with a name starting with "brian2" gets handled correctly
    submodule_logger = get_logger("brian2submodule.dummy")
    BrianLogger.initialize()
    submodule_logger.error("error message xxx")
    submodule_logger.warn("warning message xxx")
    submodule_logger.info("info message xxx")
    submodule_logger.debug("debug message xxx")
    submodule_logger.diagnostic("diagnostic message xxx")
    BrianLogger.file_handler.flush()
    # By default, only >= debug messages should show up
    assert os.path.isfile(BrianLogger.tmp_log)
    with open(BrianLogger.tmp_log, encoding="utf-8") as f:
        log_content = f.readlines()
    for level, line in zip(["error", "warning", "info", "debug"], log_content[-4:]):
        assert "submodule.dummy" in line
        # The logger name has brian2 internally prefixed, but this shouldn't show up in logs
        assert not "brian2.submodule.dummy" in line
        assert f"{level} message xxx" in line
        assert level.upper() in line

    with catch_logs() as l:
        logger.warn("warning message from Brian")
        submodule_logger.warn("warning message from submodule")
    # only the warning from Brian should be logged
    assert len(l) == 1
    assert "warning message from Brian" in l[0]


if __name__ == "__main__":
    test_file_logging()
    test_file_logging_special_characters()
    test_file_logging_multiprocessing()
    test_file_logging_multiprocessing_with_loggers()



"""
Ball and stick with Na and K channels
"""
from brian2 import *

defaultclock.dt = 0.025*ms

# Morphology
morpho = Soma(30*um)
morpho.axon = Cylinder(diameter=1*um, length=300*um, n=100)

# Channels
gL = 1e-4*siemens/cm**2
EL = -70*mV
ENa = 50*mV
ka = 6*mV
ki = 6*mV
va = -30*mV
vi = -50*mV
EK = -90*mV
vk = -20*mV
kk = 8*mV
eqs = '''
Im = gL*(EL-v)+gNa*m*h*(ENa-v)+gK*n*(EK-v) : amp/meter**2
dm/dt = (minf-m)/(0.3*ms) : 1 # simplified Na channel
dh/dt = (hinf-h)/(3*ms) : 1 # inactivation
dn/dt = (ninf-n)/(5*ms) : 1 # K+
minf = 1/(1+exp((va-v)/ka)) : 1
hinf = 1/(1+exp((v-vi)/ki)) : 1
ninf = 1/(1+exp((vk-v)/kk)) : 1
I : amp (point current)
gNa : siemens/meter**2
gK : siemens/meter**2
'''

neuron = SpatialNeuron(morphology=morpho, model=eqs,
                       Cm=1*uF/cm**2, Ri=100*ohm*cm, method='exponential_euler')
neuron.v = -65*mV
neuron.I = 0*amp
neuron.axon[30*um:60*um].gNa = 700*gL
neuron.axon[30*um:60*um].gK = 700*gL

# Monitors
mon=StateMonitor(neuron, 'v', record=True)

run(1*ms)
neuron.main.I = 0.15*nA
run(50*ms)
neuron.I = 0*amp
run(95*ms, report='text')

plot(mon.t/ms, mon.v[0]/mV, 'r')
plot(mon.t/ms, mon.v[20]/mV, 'g')
plot(mon.t/ms, mon.v[40]/mV, 'b')
plot(mon.t/ms, mon.v[60]/mV, 'k')
plot(mon.t/ms, mon.v[80]/mV, 'y')
xlabel('Time (ms)')
ylabel('v (mV)')
show()



"""
========
briandoc
========

Sphinx extension that handles docstrings in the Numpy standard format with some
brian-specific tweaks. [1]

It will:

- Convert Parameters etc. sections to field lists.
- Convert See Also section to a See also entry.
- Renumber references.
- Extract the signature from the docstring, if it can't be determined otherwise.

.. [1] https://github.com/numpy/numpy/blob/master/doc/HOWTO_DOCUMENT.rst.txt

"""

import inspect
import pydoc
import re

from docutils import nodes, statemachine
from docutils.parsers.rst import Directive, directives
from docutils.statemachine import ViewList
from sphinx.domains.c import CDomain
from sphinx.domains.python import PythonDomain, PyXRefRole
from sphinx.roles import XRefRole

from brian2.core.preferences import prefs

from .docscrape_sphinx import SphinxDocString, get_doc_object


class BrianPrefsDirective(Directive):
    """
    A sphinx 'Directive' for automatically generated documentation of Brian preferences.

    The directive takes an optional argument, the basename of the preferences
    to document. In addition, you can specify a `nolinks` option which means
    that no target links for the references are added. Do this if you document
    preferences in more then one place.

    Examples
    --------

    Document one category of preferences and generate links::

        .. document_brian_prefs:: core

    Document all preferences without generating links::

        .. document_brian_prefs::
           :nolinks:
    """

    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = True
    option_spec = {"nolinks": directives.flag, "as_file": directives.flag}
    has_content = False

    def run(self):
        # The section that should be documented
        if len(self.arguments):
            section = self.arguments[0]
        else:
            section = None
        if "as_file" in self.options:
            rawtext = prefs.as_file
            return [nodes.literal_block(text=rawtext)]
        else:
            rawtext = prefs.get_documentation(section, "nolinks" not in self.options)
            include_lines = statemachine.string2lines(rawtext, convert_whitespace=True)
            self.state_machine.insert_input(include_lines, "Brian preferences")
            return []


def brianobj_role(role, rawtext, text, lineno, inliner, options=None, content=None):
    """
    A Sphinx role, used as a wrapper for the default `py:obj` role, allowing
    us to use the simple backtick syntax for brian classes/functions without
    having to qualify the package for classes/functions that are available after
    a `from brian2 import *`, e.g `NeuronGroup`.
    Also allows to directly link to preference names using the same syntax.
    """
    if options is None:
        options = {}
    if content is None:
        content = []
    if text in prefs:
        linktext = text.replace("_", "-").replace(".", "-")
        text = f"{text} <brian-pref-{linktext}>"
        # Use sphinx's cross-reference role
        xref = XRefRole(warn_dangling=True)
        return xref("std:ref", rawtext, text, lineno, inliner, options, content)
    else:
        if text and ("~" not in text):
            try:
                # A simple class or function name
                if "." not in text:
                    module = __import__("brian2", fromlist=[str(text)])
                    imported = getattr(module, str(text), None)
                    if getattr(imported, "__module__", None):
                        text = f"~{imported.__module__}.{text}"
                        if inspect.isfunction(imported):
                            text += "()"
                # Possibly a method/classmethod/attribute name
                elif len(text.split(".")) == 2:
                    classname, attrname = text.split(".")
                    # Remove trailing parentheses (will be readded for display)
                    if attrname.endswith("()"):
                        attrname = attrname[:-2]
                    module = __import__("brian2", fromlist=[str(classname)])
                    imported = getattr(module, str(classname), None)
                    if hasattr(imported, "__module__"):
                        # Add trailing parentheses only for methods not for
                        # attributes
                        if inspect.ismethod(getattr(imported, str(attrname), None)):
                            parentheses = "()"
                        else:
                            parentheses = ""

                        text = (
                            f"{classname}.{attrname}{parentheses} "
                            f"<{imported.__module__}.{classname}.{attrname}>"
                        )

            except ImportError:
                pass
        role = "py:obj"
        py_role = PyXRefRole()
        return py_role(role, rawtext, text, lineno, inliner, options, content)


def mangle_docstrings(app, what, name, obj, options, lines, reference_offset=None):
    if reference_offset is None:
        reference_offset = [0]
    cfg = dict()
    if what == "module":
        # Strip top title
        title_re = re.compile(r"^\s*[#*=]{4,}\n[a-z0-9 -]+\n[#*=]{4,}\s*", re.I | re.S)
        lines[:] = title_re.sub("", "\n".join(lines)).split("\n")
        exported_members = getattr(obj, "__all__", None)
        if exported_members:
            lines.append("*Exported members:* ")
            # do not print more than 25 members
            lines.append(", ".join([f"`{member}`" for member in exported_members[:25]]))
            if len(exported_members) > 25:
                lines.append(f"... ({int(len(exported_members) - 25)} more members)")

            lines.append("")
    else:
        doc = get_doc_object(obj, what, "\n".join(lines), name=name, config=cfg)
        lines[:] = str(doc).split("\n")

    # replace reference numbers so that there are no duplicates
    references = []
    for line in lines:
        line = line.strip()
        m = re.match(r"^.. \[([a-z0-9_.-])\]", line, re.I)
        if m:
            references.append(m.group(1))

    # start renaming from the longest string, to avoid overwriting parts
    references.sort(key=lambda x: -len(x))
    if references:
        for i in range(len(lines)):
            for r in references:
                if re.match(r"^\d+$", r):
                    new_r = f"R{int(reference_offset[0] + int(r))}"
                else:
                    new_r = f"{r}{int(reference_offset[0])}"
                lines[i] = lines[i].replace(f"[{r}]_", f"[{new_r}]_")
                lines[i] = lines[i].replace(f".. [{r}]", f".. [{new_r}]")

    reference_offset[0] += len(references)


def mangle_signature(app, what, name, obj, options, sig, retann):
    # Do not try to inspect classes that don't define `__init__`
    if inspect.isclass(obj) and (
        not hasattr(obj, "__init__")
        or "initializes x; see " in pydoc.getdoc(obj.__init__)
    ):
        return "", ""

    if not (callable(obj) or hasattr(obj, "__argspec_is_invalid_")):
        return
    if not hasattr(obj, "__doc__"):
        return

    doc = SphinxDocString(pydoc.getdoc(obj))
    if doc["Signature"]:
        sig = re.sub("^[^(]*", "", doc["Signature"])
        return sig, ""


def setup(app, get_doc_object_=get_doc_object):
    global get_doc_object
    get_doc_object = get_doc_object_

    app.connect("autodoc-process-docstring", mangle_docstrings)
    app.connect("autodoc-process-signature", mangle_signature)

    # Extra mangling domains
    app.add_domain(NumpyPythonDomain)
    app.add_domain(NumpyCDomain)

    directives.register_directive("document_brian_prefs", BrianPrefsDirective)

    # provide the brianobj role with a link to the Python domain
    app.add_role("brianobj", brianobj_role)


# ------------------------------------------------------------------------------
# Docstring-mangling domains
# ------------------------------------------------------------------------------


class ManglingDomainBase:
    directive_mangling_map = {}

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.wrap_mangling_directives()

    def wrap_mangling_directives(self):
        for name, objtype in self.directive_mangling_map.items():
            self.directives[name] = wrap_mangling_directive(
                self.directives[name], objtype
            )


class NumpyPythonDomain(ManglingDomainBase, PythonDomain):
    name = "np"
    directive_mangling_map = {
        "function": "function",
        "class": "class",
        "exception": "class",
        "method": "function",
        "classmethod": "function",
        "staticmethod": "function",
        "attribute": "attribute",
    }


class NumpyCDomain(ManglingDomainBase, CDomain):
    name = "np-c"
    directive_mangling_map = {
        "function": "function",
        "member": "attribute",
        "macro": "function",
        "type": "class",
        "var": "object",
    }


def wrap_mangling_directive(base_directive, objtype):
    class directive(base_directive):
        def run(self):
            env = self.state.document.settings.env

            name = None
            if self.arguments:
                m = re.match(r"^(.*\s+)?(.*?)(\(.*)?", self.arguments[0])
                name = m.group(2).strip()

            if not name:
                name = self.arguments[0]

            lines = list(self.content)
            mangle_docstrings(env.app, objtype, name, None, None, lines)
            self.content = ViewList(lines, self.content.parent)

            return base_directive.run(self)

    return directive



# coding=utf-8
"""
Show the improved numerical accuracy when using the `exprel` function in rate equations.

Rate equations for channel opening/closing rates often include a term of the form
:math:`\frac{x}{\exp(x) - 1}`. This term is problematic for two reasons:

* It is not defined for :math:`x = 0` (where it should equal to :math:`1` for
  continuity);
* For values :math:`x \approx 0`, there is a loss of accuracy.

For better accuracy, and to avoid issues at :math:`x = 0`, Brian provides the
function `exprel`, which is equivalent to :math:`\frac{\exp(x) - 1}{x}`, but
with better accuracy and the expected result at :math:`x = 0`. In this example,
we demonstrate the advantage of expressing a typical rate equation from the HH
model with `exprel`.
"""
from brian2 import *

# Dummy group to evaluate the rate equation at various points
eqs = '''v : volt
         # opening rate from the HH model
         alpha_simple = 0.32*(mV**-1)*(-50*mV-v)/
                        (exp((-50*mV-v)/(4*mV))-1.)/ms : Hz
         alpha_improved = 0.32*(mV**-1)*4*mV/exprel((-50*mV-v)/(4*mV))/ms : Hz'''
neuron = NeuronGroup(1000, eqs)

# Use voltage values around the problematic point
neuron.v = np.linspace(-50 - .5e-6, -50 + .5e-6, len(neuron))*mV

fig, ax = plt.subplots()
ax.plot((neuron.v + 50*mV)/nvolt, neuron.alpha_simple,
         '.', label=r'$\alpha_\mathrm{simple}$')
ax.plot((neuron.v + 50*mV)/nvolt, neuron.alpha_improved,
         'k', label=r'$\alpha_\mathrm{improved}$')
ax.legend()
ax.set(xlabel='$v$ relative to -50mV (nV)', ylabel=r'$\alpha$ (Hz)')
ax.ticklabel_format(useOffset=False)
plt.tight_layout()
plt.show()



"""
Model fitting with simulation-based inference
---------------------------------------------

In this example, a HH-type model is used to demonstrate simulation-based
inference with the sbi toolbox (https://www.mackelab.org/sbi/). It is based
on a fake current-clamp recording generated from the same model that we use
in the inference process. Two of the parameters (the maximum sodium and
potassium conductances) are considered parameters of the model.

For more details about this approach, see the references below.

To run this example, you need to install the sbi package, e.g. with::

    pip install sbi

References:

* https://www.mackelab.org/sbi
* Tejero-Cantero et al., (2020). sbi: A toolkit for simulation-based inference.
  Journal of Open Source Software, 5(52), 2505, https://doi.org/10.21105/joss.02505
"""
import matplotlib.pyplot as plt

from brian2 import *
import sbi.utils
import sbi.analysis
import sbi.inference
import torch  # PyTorch

defaultclock.dt = 0.05*ms

def simulate(params, I=1*nA, t_on=50*ms, t_total=350*ms):
    """
    Simulates the HH-model with Brian2 for parameter sets in params and the
    given input current (injection of I between t_on and t_total-t_on).

    Returns a dictionary {'t': time steps, 'v': voltage,
                          'I_inj': current, 'spike_count': spike count}.
    """
    assert t_total > 2*t_on
    t_off = t_total - t_on
    
    params = np.atleast_2d(params)
    # fixed parameters
    gleak = 10*nS
    Eleak = -70*mV
    VT = -60.0*mV
    C = 200*pF
    ENa = 53*mV
    EK = -107*mV

    # The conductance-based model
    eqs = '''
         dVm/dt = -(gNa*m**3*h*(Vm - ENa) + gK*n**4*(Vm - EK) + gleak*(Vm - Eleak) - I_inj) / C : volt
         I_inj = int(t >= t_on and t < t_off)*I : amp (shared)
         dm/dt = alpham*(1-m) - betam*m : 1
         dn/dt = alphan*(1-n) - betan*n : 1
         dh/dt = alphah*(1-h) - betah*h : 1

         alpham = (-0.32/mV) * (Vm - VT - 13.*mV) / (exp((-(Vm - VT - 13.*mV))/(4.*mV)) - 1)/ms : Hz
         betam = (0.28/mV) * (Vm - VT - 40.*mV) / (exp((Vm - VT - 40.*mV)/(5.*mV)) - 1)/ms : Hz

         alphah = 0.128 * exp(-(Vm - VT - 17.*mV) / (18.*mV))/ms : Hz
         betah = 4/(1 + exp((-(Vm - VT - 40.*mV)) / (5.*mV)))/ms : Hz

         alphan = (-0.032/mV) * (Vm - VT - 15.*mV) / (exp((-(Vm - VT - 15.*mV)) / (5.*mV)) - 1)/ms : Hz
         betan = 0.5*exp(-(Vm - VT - 10.*mV) / (40.*mV))/ms : Hz
         # The parameters to fit
         gNa : siemens (constant)
         gK : siemens (constant)
         '''
    neurons = NeuronGroup(params.shape[0], eqs, threshold='m>0.5', refractory='m>0.5',
                          method='exponential_euler', name='neurons')
    Vm_mon = StateMonitor(neurons, 'Vm', record=True, name='Vm_mon')
    spike_mon = SpikeMonitor(neurons, record=False, name='spike_mon')  #record=False → do not record times
    neurons.gNa_ = params[:, 0]*uS
    neurons.gK = params[:, 1]*uS

    neurons.Vm = 'Eleak'
    neurons.m = '1/(1 + betam/alpham)'         # Would be the solution when dm/dt = 0
    neurons.h = '1/(1 + betah/alphah)'         # Would be the solution when dh/dt = 0
    neurons.n = '1/(1 + betan/alphan)'         # Would be the solution when dn/dt = 0

    run(t_total)
    # For convenient plotting, reconstruct the current
    I_inj = ((Vm_mon.t >= t_on) & (Vm_mon.t < t_off))*I
    return dict(v=Vm_mon.Vm,
                t=Vm_mon.t,
                I_inj=I_inj,
                spike_count=spike_mon.count)


def calculate_summary_statistics(x):
    """Calculate summary statistics for results in x"""
    I_inj = x["I_inj"]
    v = x["v"]/mV
    
    spike_count = x["spike_count"]
    # Mean and standard deviation during stimulation
    v_active = v[:, I_inj > 0*nA]
    mean_active = np.mean(v_active, axis=1)
    std_active = np.std(v_active, axis=1)
    # Height of action potential peaks
    max_v = np.max(v_active, axis=1)
    
    # concatenation of summary statistics
    sum_stats = np.vstack((spike_count, mean_active, std_active, max_v))

    return sum_stats.T


def simulation_wrapper(params):
    """
    Returns summary statistics from conductance values in `params`.
    Summarizes the output of the simulation and converts it to `torch.Tensor`.
    """
    obs = simulate(params)
    summstats = torch.as_tensor(calculate_summary_statistics(obs))
    return summstats.to(torch.float32)


if __name__ == '__main__':
    # Define prior distribution over parameters
    prior_min = [.5, 1e-4]  # (gNa, gK) in µS
    prior_max = [80.,15.]
    prior = sbi.utils.torchutils.BoxUniform(low=torch.as_tensor(prior_min),
                                            high=torch.as_tensor(prior_max))

    # Simulate samples from the prior distribution
    theta = prior.sample((10_000,))
    print('Simulating samples from prior simulation... ', end='')
    stats = simulation_wrapper(theta.numpy())
    print('done.')

    # Train inference network
    density_estimator_build_fun = sbi.utils.posterior_nn(model='mdn')
    inference = sbi.inference.SNPE(prior,
                                   density_estimator=density_estimator_build_fun)
    print('Training inference network... ')
    inference.append_simulations(theta, stats).train()
    posterior = inference.build_posterior()

    # true parameters for real ground truth data
    true_params = np.array([[32., 1.]])
    true_data = simulate(true_params)
    t = true_data['t']
    I_inj = true_data['I_inj']
    v = true_data['v']
    xo = calculate_summary_statistics(true_data)
    print("The true summary statistics are:  ", xo)

    # Plot estimated posterior distribution
    samples = posterior.sample((1000,), x=xo, show_progress_bars=False)
    labels_params = [r'$\overline{g}_{Na}$', r'$\overline{g}_{K}$']
    sbi.analysis.pairplot(samples,
                          limits=[[.5, 80], [1e-4, 15.]],
                          ticks=[[.5, 80], [1e-4, 15.]],
                          figsize=(4, 4),
                          points=true_params, labels=labels_params,
                          points_offdiag={'markersize': 6},
                          points_colors=['r'])
    plt.tight_layout()

    # Draw a single sample from the posterior and convert to numpy for plotting.
    posterior_sample = posterior.sample((1,), x=xo,
                                        show_progress_bars=False).numpy()
    x = simulate(posterior_sample)

    # plot observation and sample
    fig, ax = plt.subplots(figsize=(8, 4))
    ax.plot(t/ms, v[0, :]/mV, lw=2, label='observation')
    ax.plot(t/ms, x['v'][0, :]/mV, '--', lw=2, label='posterior sample')
    ax.legend()
    ax.set(xlabel='time (ms)', ylabel='voltage (mV)')
    plt.show()



"""
Implementation of the namespace system, used to resolve the identifiers in
model equations of `NeuronGroup` and `Synapses`
"""

import collections
import inspect
import itertools

from brian2.core.functions import DEFAULT_CONSTANTS, DEFAULT_FUNCTIONS
from brian2.units.fundamentalunits import (
    additional_unit_register,
    standard_unit_register,
)
from brian2.units.stdunits import stdunits
from brian2.utils.logger import get_logger

__all__ = [
    "get_local_namespace",
    "DEFAULT_FUNCTIONS",
    "DEFAULT_UNITS",
    "DEFAULT_CONSTANTS",
]

logger = get_logger(__name__)


def get_local_namespace(level):
    """
    Get the surrounding namespace.

    Parameters
    ----------
    level : int, optional
        How far to go back to get the locals/globals. Each function/method
        call should add ``1`` to this argument, functions/method with a
        decorator have to add ``2``.

    Returns
    -------
    namespace : dict
        The locals and globals at the given depth of the stack frame.
    """
    # Get the locals and globals from the stack frame
    frame = inspect.currentframe()
    for _ in range(level + 1):
        frame = frame.f_back
    # We return the full stack here, even if it contains a lot of stuff we are
    # not interested in -- it is cheaper to later raise an error when we find
    # a specific object with an incorrect type instead of going through this big
    # list now to check the types of all objects
    return dict(itertools.chain(frame.f_globals.items(), frame.f_locals.items()))


def _get_default_unit_namespace():
    """
    Return the namespace that is used by default for looking up units when
    defining equations. Contains all registered units and everything from
    `brian2.units.stdunits` (ms, mV, nS, etc.).

    Returns
    -------
    namespace : dict
        The unit namespace
    """
    namespace = collections.OrderedDict(standard_unit_register.units)
    namespace.update(stdunits)
    # Include all "simple" units from additional_units, i.e. units like mliter
    # but not "newton * metre"
    namespace.update(
        {
            name: unit
            for name, unit in additional_unit_register.units.items()
            if not unit.iscompound
        }
    )
    return namespace


DEFAULT_UNITS = _get_default_unit_namespace()



"""
Implementation of `BinomialFunction`
"""

import numpy as np

from brian2.core.base import Nameable
from brian2.core.functions import DEFAULT_FUNCTIONS, Function
from brian2.units.fundamentalunits import check_units
from brian2.utils.stringtools import replace

__all__ = ["BinomialFunction"]


def _pre_calc_constants_approximated(n, p):
    loc = n * p
    scale = np.sqrt(n * p * (1 - p))
    return loc, scale


def _pre_calc_constants(n, p):
    reverse = p > 0.5
    if reverse:
        P = 1.0 - p
    else:
        P = p
    q = 1.0 - P
    qn = np.exp(n * np.log(q))
    bound = min(n, n * P + 10.0 * np.sqrt(n * P * q + 1))
    return reverse, q, P, qn, bound


def _generate_cython_code(n, p, use_normal, name):
    # Cython implementation
    # Inversion transform sampling
    if use_normal:
        loc, scale = _pre_calc_constants_approximated(n, p)
        cython_code = """
        cdef float %NAME%(const int vectorisation_idx):
            return _randn(vectorisation_idx) * %SCALE% + %LOC%
        """
        cython_code = replace(
            cython_code,
            {"%SCALE%": f"{scale:.15f}", "%LOC%": f"{loc:.15f}", "%NAME%": name},
        )
        dependencies = {"_randn": DEFAULT_FUNCTIONS["randn"]}
    else:
        reverse, q, P, qn, bound = _pre_calc_constants(n, p)
        # The following code is an almost exact copy of numpy's
        # rk_binomial_inversion function
        # (numpy/random/mtrand/distributions.c)
        cython_code = """
        cdef long %NAME%(const int vectorisation_idx):
            cdef long X = 0
            cdef double px = %QN%
            cdef double U = _rand(vectorisation_idx)
            while U > px:
                X += 1
                if X > %BOUND%:
                    X = 0
                    px = %QN%
                    U = _rand(vectorisation_idx)
                else:
                    U -= px
                    px = ((%N%-X+1) * %P% * px)/(X*%Q%)
            return %RETURN_VALUE%
        """
        cython_code = replace(
            cython_code,
            {
                "%N%": f"{int(n)}",
                "%P%": f"{p:.15f}",
                "%Q%": f"{q:.15f}",
                "%QN%": f"{qn:.15f}",
                "%BOUND%": f"{bound:.15f}",
                "%RETURN_VALUE%": f"{int(n)}-X" if reverse else "X",
                "%NAME%": name,
            },
        )
        dependencies = {"_rand": DEFAULT_FUNCTIONS["rand"]}

    return cython_code, dependencies


def _generate_cpp_code(n, p, use_normal, name):
    # C++ implementation
    # Inversion transform sampling
    if use_normal:
        loc, scale = _pre_calc_constants_approximated(n, p)
        cpp_code = """
        float %NAME%(const int vectorisation_idx)
        {
            return _randn(vectorisation_idx) * %SCALE% + %LOC%;
        }
        """
        cpp_code = replace(
            cpp_code,
            {"%SCALE%": f"{scale:.15f}", "%LOC%": f"{loc:.15f}", "%NAME%": name},
        )
        dependencies = {"_randn": DEFAULT_FUNCTIONS["randn"]}
    else:
        reverse, q, P, qn, bound = _pre_calc_constants(n, p)
        # The following code is an almost exact copy of numpy's
        # rk_binomial_inversion function
        # (numpy/random/mtrand/distributions.c)
        cpp_code = """
        long %NAME%(const int vectorisation_idx)
        {
            long X = 0;
            double px = %QN%;
            double U = _rand(vectorisation_idx);
            while (U > px)
            {
                X++;
                if (X > %BOUND%)
                {
                    X = 0;
                    px = %QN%;
                    U = _rand(vectorisation_idx);
                } else
                {
                    U -= px;
                    px = ((%N%-X+1) * %P% * px)/(X*%Q%);
                }
            }
            return %RETURN_VALUE%;
        }
        """
        cpp_code = replace(
            cpp_code,
            {
                "%N%": f"{int(n)}",
                "%P%": f"{P:.15f}",
                "%Q%": f"{q:.15f}",
                "%QN%": f"{qn:.15f}",
                "%BOUND%": f"{bound:.15f}",
                "%RETURN_VALUE%": f"{int(n)}-X" if reverse else "X",
                "%NAME%": name,
            },
        )
        dependencies = {"_rand": DEFAULT_FUNCTIONS["rand"]}

    return {"support_code": cpp_code}, dependencies


class BinomialFunction(Function, Nameable):
    r"""
    BinomialFunction(n, p, approximate=True, name='_binomial*')

    A function that generates samples from a binomial distribution.

    Parameters
    ----------
    n : int
        Number of samples
    p : float
        Probablility
    approximate : bool, optional
        Whether to approximate the binomial with a normal distribution if
        :math:`n p > 5 \wedge n (1 - p) > 5`. Defaults to ``True``.
    """

    #: Container for implementing functions for different targets
    #: This container can be extended by other codegeneration targets/devices
    #: The key has to be the name of the target, the value a function
    #: that takes three parameters (n, p, use_normal) and returns a tuple of
    #: (code, dependencies)
    implementations = {"cpp": _generate_cpp_code, "cython": _generate_cython_code}

    @check_units(n=1, p=1)
    def __init__(self, n, p, approximate=True, name="_binomial*"):
        Nameable.__init__(self, name)

        # Python implementation
        use_normal = approximate and (n * p > 5) and n * (1 - p) > 5
        if use_normal:
            loc = n * p
            scale = np.sqrt(n * p * (1 - p))

            def sample_function(vectorisation_idx):
                try:
                    N = len(vectorisation_idx)
                except TypeError:
                    N = int(vectorisation_idx)
                return np.random.normal(loc, scale, size=N)

        else:

            def sample_function(vectorisation_idx):
                try:
                    N = len(vectorisation_idx)
                except TypeError:
                    N = int(vectorisation_idx)
                return np.random.binomial(n, p, size=N)

        Function.__init__(
            self,
            pyfunc=lambda: sample_function(1),
            arg_units=[],
            return_unit=1,
            stateless=False,
            auto_vectorise=True,
        )

        self.implementations.add_implementation("numpy", sample_function)

        for target, func in BinomialFunction.implementations.items():
            code, dependencies = func(n=n, p=p, use_normal=use_normal, name=self.name)
            self.implementations.add_implementation(
                target, code, dependencies=dependencies, name=self.name
            )



#!/usr/bin/env python
"""
Reliability of spike timing
---------------------------
Adapted from Fig. 10D,E of

Brette R and E Guigon (2003). Reliability of Spike Timing Is a General Property
of Spiking Model Neurons. Neural Computation 15, 279-308.

This shows that reliability of spike timing is a generic property of spiking
neurons, even those that are not leaky.
This is a non-physiological model which can be leaky or anti-leaky depending
on the sign of the input I.

All neurons receive the same fluctuating input, scaled by a parameter p that
varies across neurons. This shows:

1. reproducibility of spike timing
2. robustness with respect to deterministic changes (parameter)
3. increased reproducibility in the fluctuation-driven regime (input crosses
   the threshold)
"""
from brian2 import *

N = 500
tau = 33*ms
taux = 20*ms
sigma = 0.02

eqs_input = '''
dx/dt = -x/taux + (2/taux)**.5*xi : 1
'''

eqs = '''
dv/dt = (v*I + 1)/tau + sigma*(2/tau)**.5*xi : 1
I = 0.5 + 3*p*B : 1
B = 2./(1 + exp(-2*x)) - 1 : 1 (shared)
p : 1
x : 1 (linked)
'''

input = NeuronGroup(1, eqs_input, method='euler')
neurons = NeuronGroup(N, eqs, threshold='v>1', reset='v=0', method='euler')
neurons.p = '1.0*i/N'
neurons.v = 'rand()'
neurons.x = linked_var(input, 'x')

M = StateMonitor(neurons, 'B', record=0)
S = SpikeMonitor(neurons)

run(1000*ms, report='text')

subplot(211)  # The input
plot(M.t/ms, M[0].B)
xticks([])
title('shared input')
subplot(212)
plot(S.t/ms, neurons.p[S.i], ',')
plot([0, 1000], [.5, .5], color='C1')
xlabel('time (ms)')
ylabel('p')
title('spiking activity')
show()



{# ITERATE_ALL { _idx } #}
{# USES_VARIABLES { N } #}
{# ALLOWS_SCALAR_WRITE #}
{% extends 'common_group.pyx' %}

{% block maincode %}
    # scalar code
    _vectorisation_idx = 1
    {{scalar_code|autoindent}}
    
    # vector code
    for _idx in range(N):
        _vectorisation_idx = _idx
        {{vector_code|autoindent}}

{% endblock %}



import numpy as np
import pytest
import sympy
from numpy.testing import assert_equal

import brian2
from brian2 import (
    DimensionMismatchError,
    Expression,
    Hz,
    Statements,
    get_dimensions,
    ms,
    mV,
    second,
    volt,
)
from brian2.core.preferences import prefs
from brian2.utils.logger import catch_logs


def sympy_equals(expr1, expr2):
    """
    Test that whether two string expressions are equal using sympy, allowing
    e.g. for ``sympy_equals("x * x", "x ** 2") == True``.
    """
    s_expr1 = sympy.nsimplify(sympy.sympify(expr1).expand())
    s_expr2 = sympy.nsimplify(sympy.sympify(expr2).expand())
    return s_expr1 == s_expr2


@pytest.mark.codegen_independent
def test_expr_creation():
    """
    Test creating expressions.
    """
    expr = Expression("v > 5 * mV")
    assert expr.code == "v > 5 * mV"
    assert (
        "v" in expr.identifiers
        and "mV" in expr.identifiers
        and not "V" in expr.identifiers
    )
    with pytest.raises(SyntaxError):
        Expression("v 5 * mV")


@pytest.mark.codegen_independent
def test_split_stochastic():
    tau = 5 * ms
    expr = Expression("(-v + I) / tau")
    # No stochastic part
    assert expr.split_stochastic() == (expr, None)

    # No non-stochastic part -- note that it should return 0 and not None
    expr = Expression("sigma*xi/tau**.5")
    non_stochastic, stochastic = expr.split_stochastic()
    assert sympy_equals(non_stochastic.code, 0)
    assert "xi" in stochastic
    assert len(stochastic) == 1
    assert sympy_equals(stochastic["xi"].code, "sigma/tau**.5")

    expr = Expression("(-v + I) / tau + sigma*xi/tau**.5")
    non_stochastic, stochastic = expr.split_stochastic()
    assert "xi" in stochastic
    assert len(stochastic) == 1
    assert sympy_equals(non_stochastic.code, "(-v + I) / tau")
    assert sympy_equals(stochastic["xi"].code, "sigma/tau**.5")

    expr = Expression("(-v + I) / tau + sigma*xi_1/tau**.5 + xi_2*sigma2/sqrt(tau_2)")
    non_stochastic, stochastic = expr.split_stochastic()
    assert set(stochastic.keys()) == {"xi_1", "xi_2"}
    assert sympy_equals(non_stochastic.code, "(-v + I) / tau")
    assert sympy_equals(stochastic["xi_1"].code, "sigma/tau**.5")
    assert sympy_equals(stochastic["xi_2"].code, "sigma2/tau_2**.5")

    expr = Expression("-v / tau + 1 / xi")
    with pytest.raises(ValueError):
        expr.split_stochastic()


@pytest.mark.codegen_independent
def test_str_repr():
    """
    Test the string representation of expressions and statements. Assumes that
    __str__ returns the complete expression/statement string and __repr__ a
    string of the form "Expression(...)" or "Statements(...)" that can be
    evaluated.
    """
    expr_string = "(v - I)/ tau"
    expr = Expression(expr_string)

    # use sympy to check for equivalence of expressions (terms may have be
    # re-arranged by sympy)
    assert sympy_equals(expr_string, str(expr))
    assert sympy_equals(expr_string, eval(repr(expr)).code)

    # Use exact string equivalence for statements
    statement_string = "v += w"
    statement = Statements(statement_string)

    assert str(statement) == "v += w"
    assert repr(statement) == "Statements('v += w')"


if __name__ == "__main__":
    test_expr_creation()
    test_split_stochastic()
    test_str_repr()



from collections import defaultdict
from weakref import ref

__all__ = ["Trackable"]


class InstanceTrackerSet(set):
    """
    A `set` of `weakref.ref` to all existing objects of a certain class.

    Should not normally be directly used.
    """

    def add(self, value):
        """
        Adds a `weakref.ref` to the ``value``
        """
        # The second argument to ref is a callback that is called with the
        # ref as argument when the object has been deleted, here we just
        # remove it from the set in that case
        wr = ref(value, self.remove)
        set.add(self, wr)

    def remove(self, value):
        """
        Removes the ``value`` (which should be a weakref) if it is in the set

        Sometimes the value will have been removed from the set by `clear`,
        so we ignore `KeyError` in this case.
        """
        try:
            set.remove(self, value)
        except KeyError:
            pass


class InstanceFollower:
    """
    Keep track of all instances of classes derived from `Trackable`

    The variable ``__instancesets__`` is a dictionary with keys which are class
    objects, and values which are `InstanceTrackerSet`, so
    ``__instanceset__[cls]`` is a set tracking all of the instances of class
    ``cls`` (or a subclass).
    """

    instance_sets = defaultdict(InstanceTrackerSet)

    def add(self, value):
        for (
            cls
        ) in (
            value.__class__.__mro__
        ):  # MRO is the Method Resolution Order which contains all the superclasses of a class
            self.instance_sets[cls].add(value)

    def get(self, cls):
        return self.instance_sets[cls]


class Trackable:
    """
    Classes derived from this will have their instances tracked.

    The `classmethod` `__instances__()` will return an `InstanceTrackerSet`
    of the instances of that class, and its subclasses.
    """

    __instancefollower__ = (
        InstanceFollower()
    )  # static property of all objects of class derived from Trackable

    def __new__(typ, *args, **kw):
        obj = object.__new__(typ)
        obj.__instancefollower__.add(obj)
        return obj

    @classmethod
    def __instances__(cls):
        return cls.__instancefollower__.get(cls)



import datetime
import os

import brian2

basedir = os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..", ".."))

# Ask for version number
print('Current version is: ' + brian2.__version__)
version = input('Enter new Brian2 version number: ').strip()

# Run script to update codemeta.json
os.system(f'python {basedir}/.codemeta/create_codemeta.py {version}')

# Set version and release date in CITATION.cff
new_lines = []
with open(f'{basedir}/CITATION.cff', 'r') as f:
    for line in f:
        if line.startswith('version: '):
            line = f"version: '{version}'\n"
        if line.startswith('date-released: '):
            line = f"date-released: '{datetime.date.today().isoformat()}'\n"
        new_lines.append(line)
with open(f'{basedir}/CITATION.cff', 'w') as f:
    f.writelines(new_lines)

# commit
os.system(f'git add {basedir}/codemeta.json && git add {basedir}/CITATION.cff && git commit -m "***** Release Brian2 {version} *****"')
# add tag
os.system(f'git tag -a -m "Release Brian2 {version}" {version}')

# print commands necessary for pushing
print('Review the last commit: ')
os.system('git show %s' % version)
print('')
print('To push, using the following command:')
print('git push --tags origin master')



"""
Tests for `SpikeGeneratorGroup`
"""

import os
import tempfile

import pytest
from numpy.testing import assert_array_equal, assert_equal

from brian2 import *
from brian2.core.network import schedule_propagation_offset
from brian2.devices.device import reinit_and_delete
from brian2.tests.utils import assert_allclose, exc_isinstance
from brian2.utils.logger import catch_logs


@pytest.mark.standalone_compatible
def test_spikegenerator_connected():
    """
    Test that `SpikeGeneratorGroup` connects properly.
    """
    G = NeuronGroup(10, "v:1")
    mon = StateMonitor(G, "v", record=True, when="end")
    indices = np.array([3, 2, 1, 1, 4, 5])
    times = np.array([6, 5, 4, 3, 3, 1]) * ms
    SG = SpikeGeneratorGroup(10, indices, times)
    S = Synapses(SG, G, on_pre="v+=1")
    S.connect(j="i")
    run(7 * ms)
    # The following neurons should not receive any spikes
    for idx in [0, 6, 7, 8, 9]:
        assert all(mon[idx].v == 0)
    offset = schedule_propagation_offset()
    # The following neurons should receive a single spike
    for idx, time in zip([2, 3, 4, 5], [5, 6, 3, 1] * ms):
        assert all(mon[idx].v[mon.t < time + offset] == 0)
        assert all(mon[idx].v[mon.t >= time + offset] == 1)
    # This neuron receives two spikes
    assert all(mon[1].v[mon.t < 3 * ms + offset] == 0)
    assert all(mon[1].v[(mon.t >= 3 * ms + offset) & (mon.t < 4 * ms + offset)] == 1)
    assert all(mon[1].v[(mon.t >= 4 * ms + offset)] == 2)


@pytest.mark.standalone_compatible
def test_spikegenerator_basic():
    """
    Basic test for `SpikeGeneratorGroup`.
    """
    indices = np.array([3, 2, 1, 1, 2, 3, 3, 2, 1])
    times = np.array([1, 4, 4, 3, 2, 4, 2, 3, 2]) * ms
    SG = SpikeGeneratorGroup(5, indices, times)
    s_mon = SpikeMonitor(SG)
    run(5 * ms)
    _compare_spikes(5, indices, times, s_mon)


@pytest.mark.standalone_compatible
def test_spikegenerator_basic_sorted():
    """
    Basic test for `SpikeGeneratorGroup` with already sorted spike events.
    """
    indices = np.array([3, 1, 2, 3, 1, 2, 1, 2, 3])
    times = np.array([1, 2, 2, 2, 3, 3, 4, 4, 4]) * ms
    SG = SpikeGeneratorGroup(5, indices, times)
    s_mon = SpikeMonitor(SG)
    run(5 * ms)
    _compare_spikes(5, indices, times, s_mon)


@pytest.mark.standalone_compatible
def test_spikegenerator_basic_sorted_with_sorted():
    """
    Basic test for `SpikeGeneratorGroup` with already sorted spike events.
    """
    indices = np.array([3, 1, 2, 3, 1, 2, 1, 2, 3])
    times = np.array([1, 2, 2, 2, 3, 3, 4, 4, 4]) * ms
    SG = SpikeGeneratorGroup(5, indices, times, sorted=True)
    s_mon = SpikeMonitor(SG)
    run(5 * ms)
    _compare_spikes(5, indices, times, s_mon)


@pytest.mark.standalone_compatible
def test_spikegenerator_period():
    """
    Basic test for `SpikeGeneratorGroup`.
    """
    indices = np.array([3, 2, 1, 1, 2, 3, 3, 2, 1])
    times = np.array([1, 4, 4, 3, 2, 4, 2, 3, 2]) * ms
    SG = SpikeGeneratorGroup(5, indices, times, period=5 * ms)

    s_mon = SpikeMonitor(SG)
    run(10 * ms)
    for idx in range(5):
        generator_spikes = sorted(
            [(idx, time) for time in times[indices == idx]]
            + [(idx, time + 5 * ms) for time in times[indices == idx]]
        )
        recorded_spikes = sorted([(idx, time) for time in s_mon.t[s_mon.i == idx]])
        assert_allclose(generator_spikes, recorded_spikes)


@pytest.mark.codegen_independent
def test_spikegenerator_extreme_period():
    """
    Basic test for `SpikeGeneratorGroup`.
    """
    indices = np.array([0, 1, 2])
    times = np.array([0, 1, 2]) * ms
    SG = SpikeGeneratorGroup(5, indices, times, period=1e6 * second)
    s_mon = SpikeMonitor(SG)
    with catch_logs() as l:
        run(10 * ms)

    assert_equal(s_mon.i, np.array([0, 1, 2]))
    assert_allclose(s_mon.t, [0, 1, 2] * ms)
    assert len(l) == 1 and l[0][1].endswith("spikegenerator_long_period")


@pytest.mark.standalone_compatible
def test_spikegenerator_period_rounding():
    # See discussion in PR #1042
    # The last spike will be considered to be in the time step *after* 1s, due
    # to the way our rounding works. Although probably not what the user
    # expects, this should therefore raise an error. In previous versions of
    # Brian, this did not raise any error but silently discarded the spike.
    with pytest.raises(ValueError):
        SpikeGeneratorGroup(
            1, [0, 0, 0], [0 * ms, 0.9 * ms, 0.99999 * ms], period=1 * ms, dt=0.1 * ms
        )
    # This should also raise a ValueError, since the last two spikes fall into
    # the same bin
    s = SpikeGeneratorGroup(
        1, [0, 0, 0], [0 * ms, 0.9 * ms, 0.96 * ms], period=1 * ms, dt=0.1 * ms
    )
    net = Network(s)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms)
    assert exc_isinstance(exc, ValueError)


def test_spikegenerator_period_repeat():
    """
    Basic test for `SpikeGeneratorGroup`.
    """
    indices = np.zeros(10)
    times = arange(0, 1, 0.1) * ms

    rec = np.rec.fromarrays([times, indices], names=["t", "i"])
    rec.sort()
    sorted_times = np.ascontiguousarray(rec.t) * 1000
    sorted_indices = np.ascontiguousarray(rec.i)
    SG = SpikeGeneratorGroup(1, indices, times, period=1 * ms)
    s_mon = SpikeMonitor(SG)
    net = Network(SG, s_mon)
    rate = PopulationRateMonitor(SG)
    for idx in range(5):
        net.run(1 * ms)
        assert (idx + 1) * len(SG.spike_time) == s_mon.num_spikes


def _compare_spikes(
    N, indices, times, recorded, start_time=0 * ms, end_time=1e100 * second
):
    for idx in range(N):
        generator_spikes = sorted([(idx, time) for time in times[indices == idx]])
        recorded_spikes = sorted(
            [
                (idx, time)
                for time in recorded.t[recorded.i == idx]
                if time >= start_time and time < end_time
            ]
        )
        assert_allclose(generator_spikes, recorded_spikes)


@pytest.mark.standalone_compatible
@pytest.mark.multiple_runs
def test_spikegenerator_change_spikes():
    indices1 = np.array([3, 2, 1, 1, 2, 3, 3, 2, 1])
    times1 = np.array([1, 4, 4, 3, 2, 4, 2, 3, 2]) * ms
    SG = SpikeGeneratorGroup(5, indices1, times1)
    s_mon = SpikeMonitor(SG)
    net = Network(SG, s_mon)
    net.run(5 * ms)

    indices2 = np.array([3, 2, 1, 1, 2, 3, 3, 2, 1, 3, 3, 3, 1, 2])
    times2 = (
        np.array([1, 4, 4, 3, 2, 4, 2, 3, 2, 4.5, 4.7, 4.8, 4.5, 4.7]) * ms + 5 * ms
    )

    SG.set_spikes(indices2, times2)
    net.run(5 * ms)

    indices3 = np.array([4, 1, 0])
    times3 = np.array([1, 3, 4]) * ms + 10 * ms

    SG.set_spikes(indices3, times3)
    net.run(5 * ms)
    device.build(direct_call=False, **device.build_options)
    _compare_spikes(5, indices1, times1, s_mon, 0 * ms, 5 * ms)
    _compare_spikes(5, indices2, times2, s_mon, 5 * ms, 10 * ms)
    _compare_spikes(5, indices3, times3, s_mon, 10 * ms)


@pytest.mark.standalone_compatible
@pytest.mark.multiple_runs
def test_spikegenerator_change_period():
    """
    Basic test for `SpikeGeneratorGroup`.
    """
    indices1 = np.array([3, 2, 1, 1, 2, 3, 3, 2, 1])
    times1 = np.array([1, 4, 4, 3, 2, 4, 2, 3, 2]) * ms
    SG = SpikeGeneratorGroup(5, indices1, times1, period=5 * ms)
    s_mon = SpikeMonitor(SG)
    net = Network(SG, s_mon)
    net.run(10 * ms)

    indices2 = np.array([3, 2, 1, 1, 2, 3, 3, 2, 1, 3, 3, 3, 1, 2])
    times2 = (
        np.array([1, 4, 4, 3, 2, 4, 2, 3, 2, 4.5, 4.7, 4.8, 4.5, 4.7]) * ms + 10 * ms
    )

    SG.set_spikes(indices2, times2)
    net.run(10 * ms)  # period should no longer be in effect
    device.build(direct_call=False, **device.build_options)

    _compare_spikes(
        5,
        np.hstack([indices1, indices1]),
        np.hstack([times1, times1 + 5 * ms]),
        s_mon,
        0 * ms,
        10 * ms,
    )
    _compare_spikes(5, indices2, times2, s_mon, 10 * ms)


@pytest.mark.codegen_independent
def test_spikegenerator_incorrect_values():
    with pytest.raises(TypeError):
        SpikeGeneratorGroup(0, [], [] * second)
    # Floating point value for N
    with pytest.raises(TypeError):
        SpikeGeneratorGroup(1.5, [], [] * second)
    # Negative index
    with pytest.raises(ValueError):
        SpikeGeneratorGroup(5, [0, 3, -1], [0, 1, 2] * ms)
    # Too high index
    with pytest.raises(ValueError):
        SpikeGeneratorGroup(5, [0, 5, 1], [0, 1, 2] * ms)
    # Negative time
    with pytest.raises(ValueError):
        SpikeGeneratorGroup(5, [0, 1, 2], [0, -1, 2] * ms)


@pytest.mark.codegen_independent
def test_spikegenerator_incorrect_period():
    """
    Test that you cannot provide incorrect period arguments or combine
    inconsistent period and dt arguments.
    """
    # Period is negative
    with pytest.raises(ValueError):
        SpikeGeneratorGroup(1, [], [] * second, period=-1 * ms)

    # Period is smaller than the highest spike time
    with pytest.raises(ValueError):
        SpikeGeneratorGroup(1, [0], [2] * ms, period=1 * ms)
    # Period is not an integer multiple of dt
    SG = SpikeGeneratorGroup(1, [], [] * second, period=1.25 * ms, dt=0.1 * ms)
    net = Network(SG)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms)
    assert exc_isinstance(exc, NotImplementedError)

    SG = SpikeGeneratorGroup(1, [], [] * second, period=0.101 * ms, dt=0.1 * ms)
    net = Network(SG)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms)
    assert exc_isinstance(exc, NotImplementedError)

    SG = SpikeGeneratorGroup(1, [], [] * second, period=3.333 * ms, dt=0.1 * ms)
    net = Network(SG)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms)
    assert exc_isinstance(exc, NotImplementedError)

    # This should not raise an error (see #1041)
    SG = SpikeGeneratorGroup(1, [], [] * ms, period=150 * ms, dt=0.1 * ms)
    net = Network(SG)
    net.run(0 * ms)

    # Period is smaller than dt
    SG = SpikeGeneratorGroup(1, [], [] * second, period=1 * ms, dt=2 * ms)
    net = Network(SG)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms)
    assert exc_isinstance(exc, ValueError)


def test_spikegenerator_rounding():
    # all spikes should fall into the first time bin
    indices = np.arange(100)
    times = np.linspace(0, 0.1, 100, endpoint=False) * ms
    SG = SpikeGeneratorGroup(100, indices, times, dt=0.1 * ms)
    mon = SpikeMonitor(SG)
    net = Network(SG, mon)
    net.run(0.1 * ms)
    assert_equal(mon.count, np.ones(100))

    # all spikes should fall in separate bins
    dt = 0.1 * ms
    indices = np.zeros(10000)
    times = np.arange(10000) * dt
    SG = SpikeGeneratorGroup(1, indices, times, dt=dt)
    target = NeuronGroup(
        1, "count : 1", threshold="True", reset="count=0"
    )  # set count to zero at every time step
    syn = Synapses(SG, target, on_pre="count+=1")
    syn.connect()
    mon = StateMonitor(target, "count", record=0, when="end")
    net = Network(SG, target, syn, mon)
    # change the schedule so that resets are processed before synapses
    net.schedule = ["start", "groups", "thresholds", "resets", "synapses", "end"]
    net.run(10000 * dt)
    assert_equal(mon[0].count, np.ones(10000))


@pytest.mark.standalone_compatible
@pytest.mark.long
def test_spikegenerator_rounding_long():
    # all spikes should fall in separate bins
    dt = 0.1 * ms
    N = 1000000
    indices = np.zeros(N)
    times = np.arange(N) * dt
    SG = SpikeGeneratorGroup(1, indices, times, dt=dt)
    target = NeuronGroup(1, "count : 1")
    syn = Synapses(SG, target, on_pre="count+=1")
    syn.connect()
    spikes = SpikeMonitor(SG)
    mon = StateMonitor(target, "count", record=0, when="end")
    run(N * dt, report="text")
    assert spikes.count[0] == N, f"expected {int(N)} spikes, got {int(spikes.count[0])}"
    assert all(np.diff(mon[0].count[:]) == 1)


@pytest.mark.standalone_compatible
@pytest.mark.long
def test_spikegenerator_rounding_period():
    # all spikes should fall in separate bins
    dt = 0.1 * ms
    N = 100
    repeats = 10000
    indices = np.zeros(N)
    times = np.arange(N) * dt
    SG = SpikeGeneratorGroup(1, indices, times, dt=dt, period=N * dt)
    target = NeuronGroup(1, "count : 1")
    syn = Synapses(SG, target, on_pre="count+=1")
    syn.connect()
    spikes = SpikeMonitor(SG)
    mon = StateMonitor(target, "count", record=0, when="end")
    run(N * repeats * dt, report="text")
    # print np.int_(np.round(spikes.t/dt))
    assert_equal(spikes.count[0], N * repeats)
    assert all(np.diff(mon[0].count[:]) == 1)


@pytest.mark.codegen_independent
def test_spikegenerator_multiple_spikes_per_bin():
    # Multiple spikes per bin are of course fine if they don't belong to the
    # same neuron
    SG = SpikeGeneratorGroup(2, [0, 1], [0, 0.05] * ms, dt=0.1 * ms)
    net = Network(SG)
    net.run(0 * ms)

    # This should raise an error
    SG = SpikeGeneratorGroup(2, [0, 0], [0, 0.05] * ms, dt=0.1 * ms)
    net = Network(SG)
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms)
    print(exc.value.__cause__)
    assert exc_isinstance(exc, ValueError)

    # More complicated scenario where dt changes between runs
    defaultclock.dt = 0.1 * ms
    SG = SpikeGeneratorGroup(2, [0, 0], [0.05, 0.15] * ms)
    net = Network(SG)
    net.run(0 * ms)  # all is fine
    defaultclock.dt = 0.2 * ms  # Now the two spikes fall into the same bin
    with pytest.raises(BrianObjectException) as exc:
        net.run(0 * ms)
    assert exc_isinstance(exc, ValueError)


@pytest.mark.standalone_compatible
@pytest.mark.multiple_runs
def test_spikegenerator_multiple_runs():
    indices = np.zeros(5)
    times = np.arange(5) * ms
    spike_gen = SpikeGeneratorGroup(1, indices, times)  # all good
    spike_mon = SpikeMonitor(spike_gen)
    run(5 * ms)
    # Setting the same spike times again should not do anything, since they are
    # before the start of the current simulation
    spike_gen.set_spikes(indices, times)
    # however, a warning should be raised
    with catch_logs() as l:
        run(5 * ms)
        device.build(direct_call=False, **device.build_options)
    assert len(l) == 1 and l[0][1].endswith("ignored_spikes")
    assert spike_mon.num_spikes == 5


def test_spikegenerator_restore():
    # Check whether SpikeGeneratorGroup works with store/restore
    # See github issue #1084
    gen = SpikeGeneratorGroup(1, [0, 0, 0], [0, 1, 2] * ms)
    mon = SpikeMonitor(gen)
    store()
    run(3 * ms)
    assert_array_equal(mon.i, [0, 0, 0])
    assert_allclose(mon.t, [0, 1, 2] * ms)
    restore()
    run(3 * ms)
    assert_array_equal(mon.i, [0, 0, 0])
    assert_allclose(mon.t, [0, 1, 2] * ms)


if __name__ == "__main__":
    import time

    start = time.time()

    test_spikegenerator_connected()
    test_spikegenerator_basic()
    test_spikegenerator_basic_sorted()
    test_spikegenerator_basic_sorted_with_sorted()
    test_spikegenerator_period()
    test_spikegenerator_period_rounding()
    test_spikegenerator_extreme_period()
    test_spikegenerator_period_repeat()
    test_spikegenerator_change_spikes()
    test_spikegenerator_change_period()
    test_spikegenerator_incorrect_values()
    test_spikegenerator_incorrect_period()
    test_spikegenerator_rounding()
    test_spikegenerator_rounding_long()
    test_spikegenerator_rounding_period()
    test_spikegenerator_multiple_spikes_per_bin()
    test_spikegenerator_multiple_runs()
    test_spikegenerator_restore()
    print("Tests took", time.time() - start)



import os, re

curdir, _ = os.path.split(__file__)
units_fname = os.path.normpath(os.path.join(curdir, '../../../brian2/units/allunits.py'))

_siprefixes = {"y":1e-24, "z":1e-21, "a":1e-18, "f":1e-15, "p":1e-12, "n":1e-9,
               "u":1e-6, "m":1e-3, "c":1e-2, "d":1e-1, "":1, "da":1e1, "h":1e2,
               "k":1e3, "M":1e6, "G":1e9, "T":1e12, "P":1e15, "E":1e18,
               "Z":1e21, "Y":1e24}

fundamental_units = ['metre', 'meter', 'kilogram', 'second', 'amp', 'ampere', 'kelvin', 'mole', 'mol', 'candle']

#### DERIVED UNITS, from http://physics.nist.gov/cuu/Units/units.html
derived_unit_table = [
        ['radian',    'rad',  'get_or_create_dimension()'],
        ['steradian', 'sr',   'get_or_create_dimension()'],
        ['hertz',     'Hz',   'get_or_create_dimension(s= -1)'],
        ['newton',    'N',    'get_or_create_dimension(m=1, kg=1, s=-2)'],
        ['pascal',    'Pa',   'get_or_create_dimension(m= -1, kg=1, s=-2)'],
        ['joule',     'J',    'get_or_create_dimension(m=2, kg=1, s=-2)'],
        ['watt',      'W',    'get_or_create_dimension(m=2, kg=1, s=-3)'],
        ['coulomb',   'C',    'get_or_create_dimension(s=1, A=1)'],
        ['volt',      'V',    'get_or_create_dimension(m=2, kg=1, s=-3, A=-1)'],
        ['farad',     'F',    'get_or_create_dimension(m= -2, kg=-1, s=4, A=2)'],
        ['ohm',       'ohm',  'get_or_create_dimension(m=2, kg=1, s= -3, A=-2)'],
        ['siemens',   'S',    'get_or_create_dimension(m= -2, kg=-1, s=3, A=2)'],
        ['weber',     'Wb',   'get_or_create_dimension(m=2, kg=1, s=-2, A=-1)'],
        ['tesla',     'T',    'get_or_create_dimension(kg=1, s=-2, A=-1)'],
        ['henry',     'H',    'get_or_create_dimension(m=2, kg=1, s=-2, A=-2)'],
        ['lumen',     'lm',   'get_or_create_dimension(cd=1)'],
        ['lux',       'lx',   'get_or_create_dimension(m=-2, cd=1)'],
        ['becquerel', 'Bq',   'get_or_create_dimension(s=-1)'],
        ['gray',      'Gy',   'get_or_create_dimension(m=2, s=-2)'],
        ['sievert',   'Sv',   'get_or_create_dimension(m=2, s=-2)'],
        ['katal',     'kat',  'get_or_create_dimension(s=-1, mol=1)'],
        ]

## Generate derived unit objects and make a table of base units from these and the fundamental ones
base_units = fundamental_units+['kilogramme', 'gram', 'gramme', 'molar']
derived = ''
for longname, shortname, definition in derived_unit_table:
    derived += '{longname} = Unit.create({definition}, "{longname}", "{shortname}")\n'.format(
                longname=longname, shortname=shortname, definition=definition)
    base_units.append(longname)

all_units = base_units + []

definitions = '######### SCALED BASE UNITS ###########\n'

# Generate scaled units for all base units
scaled_units = []
excluded_scaled_units = set()
for _bu in base_units:
    if _bu in ['kelvin', 'kilogram', 'kilogramme']:
        continue  # do not create "mkelvin", "kkilogram" etc.
    for _k in _siprefixes.keys():
        if len(_k):
            _u = _k+_bu
            all_units.append(_u)
            definitions += '{_u} = Unit.create_scaled_unit({_bu}, "{_k}")\n'.format(
                                                        _u=_u, _bu=_bu, _k=_k)
            if _k not in ["da", "d", "c", "h"]:
                scaled_units.append(_u)
            else:
                excluded_scaled_units.add(_u)

# Generate 2nd and 3rd powers for all scaled base units
definitions += '######### SCALED BASE UNITS TO POWERS ###########\n'
powered_units = []
for bu in all_units + []:
    for i in [2, 3]:
        u = bu+str(i)
        definitions += '{u} = Unit.create(({bu}**{i}).dim, name="{u}", ' \
                       'dispname=str({bu})+"^{i}", ' \
                       'scale={bu}.scale*{i})\n'.format(u=u, bu=bu, i=i)
        all_units.append(u)
        if bu not in excluded_scaled_units:
            powered_units.append(u)

        additional_units = '''
# Current list from http://physics.nist.gov/cuu/Units/units.html, far from complete
additional_units = [
    pascal * second, newton * metre, watt / metre ** 2, joule / kelvin,
    joule / (kilogram * kelvin), joule / kilogram, watt / (metre * kelvin),
    joule / metre ** 3, volt / metre ** 3, coulomb / metre ** 3, coulomb / metre ** 2,
    farad / metre, henry / metre, joule / mole, joule / (mole * kelvin),
    coulomb / kilogram, gray / second, katal / metre ** 3,
    # We don't want liter/litre to be used as a standard unit for display, so we
    # put it here instead of in the standard units
    '''

# Scaled units for liter (we don't want liter**2, etc., so it is not included
# in the base units above; also, we don't want it to register in the standard
# units, so it is not chosen instead of meter**3 for displaying volume
# quantities)
for _bu in ['liter', 'litre']:
    all_units.append(_bu)
    for _k in _siprefixes.keys():
        _u = _k + _bu
        all_units.append(_u)
        definitions += '{_u} = Unit.create_scaled_unit({_bu}, "{_k}")\n'.format(
                                                    _u=_u, _bu=_bu, _k=_k)
        additional_units += _u + ', '
additional_units += ']'

# Add unit names to __all__
all = '''
__all__ = [
{allunits}
    "celsius"  # Dummy object raising an error
    ]
'''.format(allunits='\n'.join('    "'+u+'",' for u in all_units))

def to_definition(name, x):
    return '''
{name} = [
{items}
    ]
'''.format(name=name, items='\n'.join('    '+i+',' for i in x))

# Note that for presentation purposes, the `UnitRegistry` uses the *last*
# defined unit for a given scale. Therefore, we reserve the order of the units
# so that the standard units come first.
with open('units_template.py') as f:
    template = f.read()
units_str = template.format(
    derived=derived, all=all, definitions=definitions,
    base_units=to_definition('base_units', base_units[::-1]),
    scaled_units=to_definition('scaled_units', scaled_units[::-1]),
    powered_units=to_definition('powered_units', powered_units[::-1]),
    all_units=to_definition('all_units', all_units[::-1]),
    additional_units=additional_units,
    )

with open(units_fname, 'w') as f:
    f.write(units_str)



# Run a simple test that uses the main simulation elements and force code
# generation to use Cython
from brian2 import prefs
from brian2.tests.test_synapses import test_transmission_simple

if __name__ == '__main__':
    prefs.codegen.target = 'cython'

    test_transmission_simple()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
This code contains an adapted version of the voltage-dependent triplet STDP rule from:
Clopath et al., Connectivity reflects coding: a model of voltage-based STDP with homeostasis, Nature Neuroscience, 2010
(http://dx.doi.org/10.1038/nn.2479)

The plasticity rule is adapted for a leaky integrate & fire model in
Brian2. In particular, the filters ``v_lowpass1`` and ``v_lowpass2`` are
incremented by a constant at every post-synaptic spike time, to
compensate for the lack of an actual spike in the integrate & fire
model. Moreover, this script does not include the homeostatic
metaplasticity.

As an illustration of the Rule, we simulate a plot analogous to figure 2b in the above article, showing the frequency dependence of plasticity as measured in:
Sjöström et al., Rate, timing and cooperativity jointly determine cortical synaptic plasticity. Neuron, 2001.
We would like to note that the parameters have been chosen arbitrarily to qualitatively
reproduce the behavior of the original works, but need additional fitting.

We kindly ask to cite both articles when using the model presented below.

This code was written by Jacopo Bono, 12/2015
"""

from brian2 import *
################################################################################
# PLASTICITY MODEL
################################################################################

#### Plasticity Parameters

V_rest = -70.*mV        # resting potential
V_thresh = -50.*mV      # spiking threshold
Theta_low = V_rest      # depolarization threshold for plasticity
x_reset = 1.            # spike trace reset value
taux = 15.*ms           # spike trace time constant
A_LTD = 1.5e-4          # depression amplitude
A_LTP = 1.5e-2          # potentiation amplitude
tau_lowpass1 = 40*ms    # timeconstant for low-pass filtered voltage
tau_lowpass2 = 30*ms    # timeconstant for low-pass filtered voltage



#### Plasticity Equations


# equations executed at every timestep
Syn_model = '''
            w_ampa:1                # synaptic weight (ampa synapse)
            '''

# equations executed only when a presynaptic spike occurs
Pre_eq = '''
         g_ampa_post += w_ampa*ampa_max_cond                                                             # increment synaptic conductance
         w_minus = A_LTD*(v_lowpass1_post/mV - Theta_low/mV)*int(v_lowpass1_post/mV - Theta_low/mV > 0)  # synaptic depression
         w_ampa = clip(w_ampa-w_minus,0,w_max)                                                           # hard bounds
         '''

# equations executed only when a postsynaptic spike occurs
Post_eq = '''
          v_lowpass1 += 10*mV                                                                                        # mimics the depolarisation by a spike
          v_lowpass2 += 10*mV                                                                                        # mimics the depolarisation by a spike
          w_plus = A_LTP*x_trace_pre*(v_lowpass2_post/mV - Theta_low/mV)*int(v_lowpass2_post/mV - Theta_low/mV > 0)  # synaptic potentiation
          w_ampa = clip(w_ampa+w_plus,0,w_max)                                                                       # hard bounds
          '''

################################################################################
# I&F Parameters and equations
################################################################################

#### Neuron parameters

gleak = 30.*nS                  # leak conductance
C = 300.*pF                     # membrane capacitance
tau_AMPA = 2.*ms                # AMPA synaptic timeconstant
E_AMPA = 0.*mV                  # reversal potential AMPA

ampa_max_cond = 5.e-10*siemens  # Ampa maximal conductance
w_max = 1.                      # maximal ampa weight


#### Neuron Equations

eqs_neurons = '''
dv/dt = (gleak*(V_rest-v) + I_ext + I_syn)/C: volt      # voltage
dv_lowpass1/dt = (v-v_lowpass1)/tau_lowpass1 : volt     # low-pass filter of the voltage
dv_lowpass2/dt = (v-v_lowpass2)/tau_lowpass2 : volt     # low-pass filter of the voltage
I_ext : amp                                             # external current
I_syn = g_ampa*(E_AMPA-v): amp                          # synaptic current
dg_ampa/dt = -g_ampa/tau_AMPA : siemens                 # synaptic conductance
dx_trace/dt = -x_trace/taux :1                          # spike trace
'''



################################################################################
# Simulation
################################################################################

#### Parameters

defaultclock.dt = 100.*us                           # timestep
Nr_neurons = 2                                      # Number of neurons
rate_array = [1., 5., 10., 15., 20., 30., 50.]*Hz   # Rates
init_weight = 0.5                                   # initial synaptic weight
reps = 15                                           # Number of pairings

#### Create neuron objects

Nrns = NeuronGroup(Nr_neurons, eqs_neurons, threshold='v>V_thresh',
                   reset='v=V_rest;x_trace+=x_reset/(taux/ms)', method='euler')#

#### create Synapses

Syn = Synapses(Nrns, Nrns,
               model=Syn_model,
               on_pre=Pre_eq,
               on_post=Post_eq
               )

Syn.connect('i!=j')

#### Monitors and storage
weight_result = np.zeros((2, len(rate_array)))               # to save the final weights

#### Run

# loop over rates
for jj, rate in enumerate(rate_array):

    # Calculate interval between pairs
    pair_interval = 1./rate - 10*ms
    print('Starting simulations for %s' % rate)

    # Initial values
    Nrns.v = V_rest
    Nrns.v_lowpass1 = V_rest
    Nrns.v_lowpass2 = V_rest
    Nrns.I_ext = 0*amp
    Nrns.x_trace = 0.
    Syn.w_ampa = init_weight

    # loop over pairings
    for ii in range(reps):
        # 1st SPIKE
        Nrns.v[0] = V_thresh + 1*mV
        # 2nd SPIKE
        run(10*ms)
        Nrns.v[1] = V_thresh + 1*mV
        # run
        run(pair_interval)
        print('Pair %d out of %d' % (ii+1, reps))

    #store weight changes
    weight_result[0, jj] = 100.*Syn.w_ampa[0]/init_weight
    weight_result[1, jj] = 100.*Syn.w_ampa[1]/init_weight

################################################################################
# Plots
################################################################################

stitle = 'Pairings'
scolor = 'k'

figure(figsize=(8, 5))
plot(rate_array, weight_result[0,:], '-', linewidth=2, color=scolor)
plot(rate_array, weight_result[1,:], ':', linewidth=2, color=scolor)
xlabel('Pairing frequency [Hz]', fontsize=22)
ylabel('Normalised Weight [%]', fontsize=22)
legend(['Pre-Post', 'Post-Pre'], loc='best')
subplots_adjust(bottom=0.2, left=0.15, right=0.95, top=0.85)
title(stitle)
show()



"""
Module defining the `ImportExport` class that enables getting state variable
data in and out of groups in various formats (see `Group.get_states` and
`Group.set_states`).
"""

import abc
from abc import abstractmethod


class ImportExport(metaclass=abc.ABCMeta):
    """
    Class for registering new import/export methods (via static methods). Also
    the base class that should be extended for such methods
    (`ImportExport.export_data`, `ImportExport.import_data`, and
    `ImportExport.name` have to be overwritten).

    See Also
    --------
    VariableOwner.get_states, VariableOwner.set_states

    """

    #: A dictionary mapping import/export methods names to `ImportExport` objects
    methods = dict()

    @staticmethod
    def register(importerexporter):
        """
        Register a import/export method. Registered methods can be referred to
        via their name.

        Parameters
        ----------
        importerexporter : `ImportExport`
            The importerexporter object, e.g. an `DictImportExport`.
        """
        if not isinstance(importerexporter, ImportExport):
            t = str(type(importerexporter))
            error_msg = (
                f"Given importerexporter of type {t} does not seem to "
                "be a valid importerexporter."
            )
            raise ValueError(error_msg)
        name = importerexporter.name
        if name in ImportExport.methods:
            raise ValueError(
                f"An import/export methods with the name '{name}'"
                "has already been registered"
            )
        ImportExport.methods[name] = importerexporter

    @staticmethod
    @abstractmethod
    def export_data(group, variables):
        """
        Asbtract static export data method with two obligatory parameters.
        It should return a copy of the current state variable values. The
        returned arrays are copies of the actual arrays that store the state
        variable values, therefore changing the values in the returned
        dictionary will not affect the state variables.

        Parameters
        ----------
        group : `Group`
            Group object.
        variables : list of str
            The names of the variables to extract.
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def import_data(group, data):
        """
        Import and set state variables.

        Parameters
        ----------
        group : `Group`
            Group object.
        data : dict_like
            Data to import with variable names.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def name(self):
        """
        Abstract property giving a method name.
        """
        pass



"""
Check that various monitors work correctly.
"""

from brian2 import *
from brian2.tests.features import FeatureTest, InaccuracyError


class SpikeMonitorTest(FeatureTest):
    category = "Monitors"
    name = "SpikeMonitor"
    tags = ["NeuronGroup", "run", "SpikeMonitor"]

    def run(self):
        N = 100
        tau = 10 * ms
        eqs = """
        dv/dt = (I-v)/tau : 1
        I : 1
        """
        self.G = G = NeuronGroup(N, eqs, threshold="v>1", reset="v=0")
        G.I = linspace(0, 2, N)
        self.M = M = SpikeMonitor(G)
        run(100 * ms)

    def results(self):
        return {"i": self.M.i[:], "t": self.M.t[:]}

    compare = FeatureTest.compare_arrays


class StateMonitorTest(FeatureTest):
    category = "Monitors"
    name = "StateMonitor"
    tags = ["NeuronGroup", "run", "StateMonitor"]

    def run(self):
        N = 10
        tau = 10 * ms
        eqs = """
        dv/dt = (I-v)/tau : 1
        I : 1
        """
        self.G = G = NeuronGroup(N, eqs, threshold="v>1", reset="v=0.1")
        G.v = 0.1
        G.I = linspace(1.1, 2, N)
        self.M = M = StateMonitor(G, "v", record=True)
        run(100 * ms)

    def results(self):
        return self.M.v[:]

    compare = FeatureTest.compare_arrays



#!/usr/bin/env python3
"""
FORCE training of a Leaky IF model to mimic a sinusoid (5 Hz) oscillator

Nicola, W., Clopath, C.
Supervised learning in spiking neural networks with FORCE training
Nat Commun 8, 2208 (2017)

https://doi.org/10.1038/s41467-017-01827-3

Based on https://github.com/ModelDBRepository/190565/blob/master/CODE%20FOR%20FIGURE%202/LIFFORCESINE.m

Sebastian Schmitt, 2022
"""

from brian2 import NeuronGroup, Synapses, StateMonitor, SpikeMonitor
from brian2 import run, defaultclock, network_operation
from brian2 import ms, second, Hz
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import numpy as np

# set seed for reproducible figures
np.random.seed(1)

# decay time of synaptic kernal
td = 20*ms

# rise time of synaptic kernal
tr = 2*ms

# membrane time constant
tm = 10*ms

# refractory period
tref = 2*ms

# reset potential
vreset = -65

# peak/threshold potential
vpeak = -40

# bias
BIAS = vpeak

# integration time step
defaultclock.dt = 0.05*ms

# total duration of simulation
T = 15*second

# start of training
imin = 5*second

# end of training
icrit = 10*second

# interval of training
step = 2.5*ms

# feedback scale factor
Q = 10

# neuron-to-neuron connection scale factor
G = 0.04

# connection probability
p = 0.1

# number of neurons
N = 2000

# correlation weight matrix for RLMS
alpha = defaultclock.dt/second*0.1
Pinv = np.eye(N)*alpha

# Sinusoid oscillator
def zx(t):
    freq = 5*Hz
    return np.sin(2*np.pi*freq*t)

neurons = NeuronGroup(N,
                      """
                      dv/dt = (-v + BIAS + IPSC + E*z)/tm: 1 (unless refractory)
                      dIPSC/dt = -IPSC/tr + h : 1
                      dh/dt = -h/td : 1/second
                      dr/dt = -r/tr + hr : 1
                      dhr/dt = -hr/td : 1/second
                      BPhi : 1
                      z : 1 (shared)
                      E : 1
                      """,
                      method="euler",
                      threshold="v>=vpeak",
                      reset="v=vreset; hr += 1/(tr*td)*second",
                      refractory=tref)

# fixed feedback weights
neurons.E = (2*np.random.uniform(size=N)-1)*Q

# initial membrane voltage
neurons.v = vreset + np.random.uniform(size=N)*(30-vreset)

synapses = Synapses(neurons, neurons, "w : second", on_pre="h += w/(tr*td)")
synapses.connect()
omega = G*(np.random.normal(size=(N,N))*(np.random.uniform(size=(N,N))<p))/(np.sqrt(N)*p)
synapses.w = omega.flatten()*second

spikemon = SpikeMonitor(neurons[:20])
statemon_BPhi = StateMonitor(neurons, "BPhi", record=range(10))
statemon_z = StateMonitor(neurons, "z", record=[0])

# linear readout
@network_operation(dt=defaultclock.dt)
def readout(t):
    neurons.z = np.dot(neurons.BPhi, neurons.r)

# FORCE training
@network_operation(dt=step)
def train(t):
    global Pinv
    if t > imin and t < icrit:
        cd = Pinv@neurons.r
        err = neurons.z - zx(t)
        neurons.BPhi -= cd*err
        Pinv -= np.outer(cd,cd)/( 1 + np.dot(neurons.r, cd))

run(T, report="text")

fig, axes = plt.subplots(2,2, figsize=(10,10))
axes = axes.flatten()

axes[0].set_title("Spike raster")
axes[0].scatter(spikemon.t/second,spikemon.i, marker='|', linestyle="None", color="black", s=100)
axes[0].set_xlim((imin-2*second)/second, imin/second+2)
axes[0].set_ylim(0, len(spikemon.source))
axes[0].set_xlabel("t [s]")
axes[0].set_ylabel("Neuron")
axes[0].yaxis.set_major_locator(MaxNLocator(integer=True))

axes[1].plot(statemon_z.t/second, zx(statemon_z.t), linestyle='--', color='k')
axes[1].plot(statemon_z.t/second,statemon_z.z[0])

axes[1].set_title("Target and readout")
axes[1].annotate('RLS ON', xy=(imin/second, -1.05), xytext=(imin/second, -1.35),
            arrowprops=dict(facecolor='black', shrink=1), ha="center")
axes[1].annotate('RLS OFF', xy=(icrit/second, -1.05), xytext=(icrit/second, -1.35),
            arrowprops=dict(facecolor='black', shrink=1), ha="center")
axes[1].set_xlabel("t [s]")
axes[1].set_xlim((imin-1*second)/second, T/second)
axes[1].set_ylim(-1.4,1.1)

axes[2].set_title("Error")
axes[2].plot(statemon_z.t/second, statemon_z.z[0] - zx(statemon_z.t))
axes[2].annotate('RLS ON', xy=(imin/second, -0.15), xytext=(imin/second, -0.4),
            arrowprops=dict(facecolor='black', shrink=1), ha="center")
axes[2].annotate('RLS OFF', xy=(icrit/second, -0.15), xytext=(icrit/second, -0.4),
            arrowprops=dict(facecolor='black', shrink=1), ha="center")
axes[2].set_xlabel("t [s]")
axes[2].set_xlim((imin-1*second)/second, T/second)
axes[2].set_ylim(-1,1)

axes[3].set_title("Decoders")
for j in range(len(statemon_BPhi.record)):
    axes[3].plot(statemon_BPhi.t/second,statemon_BPhi.BPhi[j])
axes[3].set_xlim((imin-1*second)/second, T/second)
axes[3].set_xlabel("t [s]")
axes[3].set_ylim(-0.00020, 0.00015)
axes[3].set_yticklabels([])
axes[3].annotate('RLS ON', xy=(imin/second, -0.0001455), xytext=(imin/second, -0.00019),
            arrowprops=dict(facecolor='black', shrink=1), ha="center")
axes[3].annotate('RLS OFF', xy=(icrit/second, -0.0001455), xytext=(icrit/second, -0.00019),
            arrowprops=dict(facecolor='black', shrink=1), ha="center")

fig.tight_layout()
